#!/usr/bin/env python3
"""
Module WX.HTML -- Python wxWidgets Html Display Widgets
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2024 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for text display widgets.
"""

import wx
import wx.html

from plib.ui.defs import *
from plib.ui.base.html import PHtmlDisplayBase

from .app import PWxWidget, ec_scroll_style


class PHtmlDisplay(PWxWidget, wx.html.HtmlWindow, PHtmlDisplayBase):
    
    _style = wx.html.HW_SCROLLBAR_AUTO
    
    _align = True  # used by panel to determine placement
    _expand = True
    
    def __init__(self, manager, parent, html=None,
                 scrolling=False, font=None):
        
        if scrolling:
            self._style = self._style | ec_scroll_style
        wx.html.HtmlWindow.__init__(self, parent, style=self._style)
        self._html = ""  # hack because wx gives no way of retrieving the html from the widget
        PHtmlDisplayBase.__init__(self, manager, parent, html,
                                  scrolling=scrolling, font=font)
    
    def setup_scrolling(self, scrolling):
        pass  # this must be done in the constructor in wxWidgets
    
    def get_html(self):
        return self._html
    
    def set_html(self, value):
        self._html = value
        self.SetPage(value)



#!/usr/bin/env python3
"""
Module WX.GROUPBOX -- Python wxWidgets Group Box Widgets
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for group box widgets.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.groupbox import PGroupBoxBase

from .app import PWxWidget


class PGroupBox(PWxWidget, wx.Panel, PGroupBoxBase):
    
    _align = False  # used by panel to determine placement
    _expand_horiz = True
    
    def __init__(self, manager, parent, caption, contents=None,
                 geometry=None, font=None, margin=None, spacing=None):
        
        wx.Panel.__init__(self, parent, style=wx.NO_BORDER)
        self._margin = margin
        self._spacing = spacing
        self._haswidgets = False
        self._box = wx.StaticBox(self)
        self._boxsizer = wx.StaticBoxSizer(self._box, wx.VERTICAL)
        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self._sizer.Add(self._boxsizer, 1, wx.EXPAND, 0)
        PGroupBoxBase.__init__(self, manager, parent, caption, contents,
                               geometry=geometry, font=font, margin=margin, spacing=spacing)
    
    def set_caption(self, caption):
        self._box.SetLabel(caption)
    
    def set_min_size(self, width, height):
        self._sizer.SetMinSize((width, height))
    
    def set_margin(self, margin):
        # Margin is dealt with differently in wxWidgets, see above
        pass
    
    def set_spacing(self, spacing):
        # Spacing is dealt with differently in wxWidgets, see above
        pass
    
    def add_control(self, control):
        # Somewhat abbreviated version of the cruft in PPanel
        
        if self._haswidgets:
            if self._spacing is not None:
                self._boxsizer.AddSpacer(self._spacing)
        elif self._margin is not None:
            self._boxsizer.AddSpacer(self._margin)
        
        if getattr(control, '_align', None):
            proportion = 1
        else:
            proportion = 0
        
        if getattr(control, '_expand', None) or getattr(control, '_expand_horiz', None):
            flag = wx.EXPAND
        else:
            flag = 0
        
        flag |= wx.LEFT | wx.RIGHT
        border = self._margin or 0
        
        self._boxsizer.Add(control, proportion, flag, border)
        if not self._haswidgets:
            self._haswidgets = True
    
    def do_layout(self):
        if self._margin is not None:
            self._boxsizer.AddSpacer(self._margin)
        self.SetSizerAndFit(self._sizer)



#!/usr/bin/env python3
"""
Module WX.SCROLLER -- Python wxWidgets Scroller Objects
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for the panel widgets.
"""

import wx

from plib.ui.base.scroller import PScrollerBase, PScrollingWrapperBase

from .app import PWxWidget


class PWxScrollerBase(PWxWidget, wx.ScrolledWindow):
    
    def __init__(self, parent):
        wx.ScrolledWindow.__init__(self, parent)
        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetScrollRate(20, 20)
    
    def set_child(self, child):
        # Note: child widget must already have been constructed with self as parent,
        # so in wx we can never pass a child widget to the PScroller constructor
        # TODO: is there a way to fix this?
        self._sizer.Add(child, 1, wx.EXPAND, 0)
        self.SetSizer(self._sizer)


class PScroller(PWxScrollerBase, PScrollerBase):
    
    def __init__(self, manager, parent, child=None):
        PWxScrollerBase.__init__(self, parent)
        PScrollerBase.__init__(self, manager, parent, child)


class PScrollingWrapper(PWxScrollerBase, PScrollingWrapperBase):
    
    def __init__(self, manager, parent, child_spec):
        PWxScrollerBase.__init__(self, parent)
        PScrollingWrapperBase.__init__(self, manager, parent, child_spec)
    
    @property
    def spec_parent(self):
        return self  # since there's no other way to have us scroll the child window
    
    # get_layout and setup_layout are not used in wx



#!/usr/bin/env python3
"""
Module MANAGER -- UI Widget Manager
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.base.app import PManagerBase


class PWidgetManager(PManagerBase):
    
    name = None
    
    main_widget = None
    
    def __init__(self, manager):
        PManagerBase.__init__(self)
        self.manager = manager
        manager.sub_managers.append(self)
    
    @property
    def manager_type(self):
        return type(self).__name__.lower()
    
    @property
    def attrname(self,
                 fmt="{}_{}".format):
        
        return fmt(self.manager_type, self.name)



#!/usr/bin/env python3
"""
Module FORM -- UI Panel Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.defs import *
from plib.ui.widgets import widgets_from_contents

from .app import PWidgetBase


class PPanelBase(PWidgetBase):
    """Base class for 'panel' widget.
    
    The panel widget is a container for laying out
    other widgets. The layout parameter is used to
    determine whether child widgets are laid out
    horizontally or vertically, or in a grid with
    a specified number of columns or rows; the align
    parameter determines how this panel takes up space
    in its parent widget. Panels should be nestable to
    any depth desired, to enable complex layouts.
    
    Note that there is currently no checking
    done to ensure that the align value of a panel
    makes sense when combined with the alignment of
    its sibling widgets and the layout of its parent
    widget (for example, an ALIGN_TOP or ALIGN_BOTTOM
    panel with ALIGN_LEFT or ALIGN_RIGHT siblings
    inside a LAYOUT_HORIZONTAL parent panel). It is
    up to the caller to ensure that the combined
    parameters make sense. (This also applies to the
    order in which child widgets are added -- for
    example, an ALIGN_LEFT widget must be added
    before an ALIGN_JUST widget in a horizontal
    layout, or the ALIGN_LEFT widget will be on the
    right.)
    """
    
    colcount = None
    rowcount = None
    
    contents = None
    
    def __init__(self, manager, parent, align=ALIGN_JUST, layout=LAYOUT_VERTICAL, contents=None,
                 style=PANEL_PLAIN, margin=None, spacing=None):
        
        PWidgetBase.__init__(self, manager, parent)
        
        if isinstance(layout, tuple):
            layout, num = layout
            if layout == LAYOUT_COLGRID:
                self.colcount = num
            elif layout == LAYOUT_ROWGRID:
                self.rowcount = num
            else:
                raise ValueError("Invalid layout value: {}".format(layout))
        
        self.layout = layout
        self.align = align
        self.style = style
        
        if (margin is not None):
            self.set_margin(margin)
        if (spacing is not None):
            self.set_spacing(spacing)
        
        # Create child panels, if any, and do layout if children were created
        self.create_contents(contents)
        self.do_layout()
    
    def add_widget(self, widget):
        # Derived classes may override to do any necessary
        # adjustment when a widget is added.
        pass
    
    def create_contents(self, contents):
        if self.contents is None:
            self.contents = []
        if contents:
            self.contents.extend(widgets_from_contents(self.manager, self, contents, self.add_widget))
    
    def do_layout(self):
        # Derived classes should override to do any
        # necessary creation of physical layout objects
        # after all child panels are created.
        pass
    
    def remove_widget(self, widget):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def set_margin(self, margin):
        """Derived classes must implement.
        """
        raise NotImplementedError
    
    def set_spacing(self, spacing):
        """Derived classes must implement.
        """
        raise NotImplementedError



#!/usr/bin/env python3
"""
Sub-Package UI.TOOLKITS of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This sub-package contains sub-sub-packages for each of the supported
UI toolkits.
"""

import sys
import os
import tempfile

from plib.stdlib.builtins import first
from plib.stdlib.proc import process_call


class ToolkitError(Exception):
    pass


DETECT_TEMPLATE = """# Toolkit detection temporary script

import sys

try:
    import {import_name}
except ImportError:
    sys.exit(1)
else:
    sys.exit(0)
"""


class _ToolkitManager(object):
    
    supported_toolkits = (
        'pyside2',
        'qt5',
        'wx',
    )
    
    def __init__(self):
        # We don't auto-initialize because we don't want to go through the
        # auto-detection process unless we have to; the application can call
        # set_toolkit before any other plib.ui modules are imported to avoid
        # triggering the auto-detection machinery
        
        self.available_toolkits = []
        self.toolkit = None
        
        env_var = os.getenv('UI_TOOLKIT')
        if env_var is not None:
            self.set_toolkit(env_var)
    
    def init_toolkit(self):
        self.available_toolkits[:] = self.detect_toolkits()
        self.toolkit = first(t for t in self.supported_toolkits if t in self.available_toolkits)
    
    import_map = {
        'pyside2': "PySide2",
        'qt5': "PyQt5",
    }
    
    script_template = DETECT_TEMPLATE
    script_filename = "toolkit_detect.py"
    
    script_dir = None
    
    def is_detected(self, toolkit):
        import_name = self.import_map.get(toolkit, toolkit)
        script_data = self.script_template.format(**locals())
        filename = os.path.join(self.script_dir, self.script_filename)
        
        exitcode = -1
        try:
            with open(filename, 'w') as f:
                f.write(script_data)
            exitcode, _ = process_call([sys.executable, filename])
        finally:
            if os.path.isfile(filename):
                os.remove(filename)
        
        return (exitcode == 0)
    
    def detect_toolkits(self):
        # We have to go through all this because having multiple toolkit modules
        # imported in the same interpreter process can cause errors, so we have
        # to test for each of them in a separate process. This makes the detection
        # machinery somewhat heavyweight, which is why we avoid using it unless we
        # have to.
        
        detected = []
        
        self.script_dir = tempfile.mkdtemp()
        try:
            for toolkit in self.supported_toolkits:
                if self.is_detected(toolkit):
                    detected.append(toolkit)
        finally:
            os.rmdir(self.script_dir)
            del self.script_dir
        
        return detected
    
    def get_toolkit(self):
        if not self.toolkit:
            print("Need to initialize toolkits")
            self.init_toolkit()
        return self.toolkit
    
    def set_toolkit(self, toolkit):
        if toolkit not in self.supported_toolkits:
            raise ToolkitError("Toolkit not supported: {}".format(toolkit))
        # We don't check for availability here, the caller is responsible for
        # handling any errors if the specified toolkit is not available
        print("Setting toolkit to", toolkit)
        self.toolkit = toolkit


_manager = _ToolkitManager()

supported_toolkits = _manager.supported_toolkits
available_toolkits = _manager.available_toolkits

get_toolkit = _manager.get_toolkit
set_toolkit = _manager.set_toolkit



#!/usr/bin/env python3
"""
Module WX.APP -- Python wxWidgets Application Objects
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI application objects.
"""

from abc import ABCMeta
import functools
import string

import wx
import wx.adv
import wx.dataview
import wx.grid
from wx.lib.evtmgr import eventManager
from wx.lib.newevent import NewEvent

from plib.ui.defs import *
from plib.ui.base.app import (
    PMessageBoxBase, PFileDialogBase, PAboutDialogBase, PProgressDialogBase,
    PTopWindowBase, PApplicationBase
)


# Style constant for edit controls that are scrolling

ec_scroll_style = wx.VSCROLL | wx.HSCROLL | wx.TE_DONTWRAP

# FIXME: These are only good for buttons and menus, and the toolbar ones
# in wx.ArtProvider don't match all the ids! WTF!@#$%^&*()
# FIXME: Even the menu ones don't all work properly
stock_ids = {
    ACTION_FILE_NEW: wx.ID_NEW,
    ACTION_FILE_OPEN: wx.ID_OPEN,
    ACTION_FILE_SAVE: wx.ID_SAVE,
    ACTION_FILE_SAVEAS: wx.ID_SAVEAS,
    ACTION_FILE_CLOSE: wx.ID_CLOSE,
    #ACTION_EDIT: wx.ID_EDIT,
    ACTION_EDIT_UNDO: wx.ID_UNDO,
    ACTION_EDIT_REDO: wx.ID_REDO,
    ACTION_EDIT_CUT: wx.ID_CUT,
    ACTION_EDIT_COPY: wx.ID_COPY,
    ACTION_EDIT_PASTE: wx.ID_PASTE,
    ACTION_EDIT_DELETE: wx.ID_DELETE,
    #ACTION_EDIT_SELECTALL: wx.ID_SELECTALL,
    #ACTION_REFRESH: wx.ID_REFRESH,
    #ACTION_ADD: wx.ID_ADD,
    #ACTION_REMOVE: wx.ID_REMOVE,
    #ACTION_APPLY: wx.ID_APPLY,
    #ACTION_OK: wx.ID_OK,
    #ACTION_CANCEL: wx.ID_CANCEL,
    #ACTION_PREFS: wx.ID_PREFERENCES,
    ACTION_ABOUT: wx.ID_ABOUT,
    ACTION_EXIT: wx.ID_EXIT
}

align_map = {
    ALIGN_LEFT: wx.ALIGN_LEFT,
    ALIGN_CENTER: wx.ALIGN_CENTER_HORIZONTAL,
    ALIGN_RIGHT: wx.ALIGN_RIGHT,
}

icon_map = {
    MBOX_INFO: wx.ICON_INFORMATION,
    MBOX_WARN: wx.ICON_EXCLAMATION,
    MBOX_ERROR: wx.ICON_ERROR,
    MBOX_QUERY: wx.ICON_QUESTION
}

font_families = {
    wx.FONTFAMILY_ROMAN: ["Times New Roman"],
    wx.FONTFAMILY_SWISS: ["Arial", "Verdana"],
    wx.FONTFAMILY_MODERN: ["Courier New"]
}

color_map = {
    COLOR_BLACK: wx.BLACK,  # wx.Colour(0, 0, 0)
    COLOR_DARKRED: wx.Colour(63, 0, 0),
    COLOR_RED: wx.RED,  # wx.Colour(127, 0, 0)
    COLOR_DARKGREEN: wx.Colour(0, 63, 0),
    COLOR_GREEN: wx.GREEN,  # wx.Colour(0, 127, 0)
    COLOR_LIGHTGREEN: wx.Colour(0, 255, 0),
    COLOR_DARKBLUE: wx.Colour(0, 0, 63),
    COLOR_BLUE: wx.BLUE,  # wx.Colour(0, 0, 127)
    COLOR_LIGHTBLUE: wx.Colour(0, 0, 255),
    COLOR_YELLOW: wx.YELLOW,  # wx.Colour(255, 255, 0)
    COLOR_MAGENTA: wx.Colour(255, 0, 255),
    COLOR_CYAN: wx.CYAN,  # wx.Colour(0, 255, 255)
    COLOR_DARKGRAY: wx.Colour(63, 63, 63),
    COLOR_GRAY: wx.Colour(127, 127, 127),
    COLOR_LIGHTGRAY: wx.LIGHT_GREY,  # wx.Colour(191, 191, 191)
    COLOR_WHITE: wx.WHITE,  # wx.Colour(255, 255, 255)
}

# Wx doesn't provide any standard shortcuts, so we have to
# roll our own; this is our best guess at what is common to
# the various possible platforms

accel_map = {
    ACTION_FILE_NEW: "Ctrl+N",
    ACTION_FILE_OPEN: "Ctrl+O",
    ACTION_FILE_SAVE: "Ctrl+S",
    ACTION_FILE_SAVEAS: "Ctrl+Shift+S",
    ACTION_FILE_CLOSE: "Ctrl+W",
    ACTION_EDIT_UNDO: "Ctrl+Z",
    ACTION_EDIT_REDO: "Ctrl+Shift+Z",
    ACTION_EDIT_CUT: "Ctrl+X",
    ACTION_EDIT_COPY: "Ctrl+C",
    ACTION_EDIT_PASTE: "Ctrl+V",
    ACTION_EDIT_DELETE: "Ctrl+D",
    ACTION_EDIT_SELECTALL: "Ctrl+A",
    ACTION_EDIT_SELECTNONE: "Ctrl+Shift+A",
    #ACTION_EDIT_OVERWRITE: "Ins",  # this already seems to be built in to wx
    #ACTION_EDIT_CLEAR: ,
    #ACTION_VIEW: ,
    #ACTION_EDIT: ,
    #ACTION_OK: ,
    #ACTION_CANCEL: ,
    ACTION_REFRESH: "F5",
    #ACTION_ADD: ,
    #ACTION_REMOVE: ,
    #ACTION_APPLY: ,
    #ACTION_COMMIT: ,
    #ACTION_ROLLBACK: ,
    #ACTION_PREFS: ,
    ACTION_ABOUT: "Alt+B",
    ACTION_ABOUT_TOOLKIT: "Alt+T",
    ACTION_EXIT: "Ctrl+Q",
}

keycode_map = {
    "F5": wx.WXK_F5,
}

control_key_map = {
    "Ctrl": wx.ACCEL_CTRL,
    "Shift": wx.ACCEL_SHIFT,
    "Alt": wx.ACCEL_ALT
}


def parse_accel(s,
                sep="+", cmap=control_key_map,
                norm=wx.ACCEL_NORMAL):
    
    if isinstance(s, wx.KeyCode):
        return s
    keys = s.split(sep)
    key = keys[-1]
    keycode = keycode_map.get(key)
    if not keycode:
        keycode = ord(key)
    flags = norm
    for c in keys[:-1]:
        flags |= cmap[c]
    return (flags, keycode)


# Define our own custom events
QueryCloseEvent, EVT_QUERYCLOSE = NewEvent()
ClosingEvent, EVT_CLOSING = NewEvent()
ShownEvent, EVT_SHOWN = NewEvent()
HiddenEvent, EVT_HIDDEN = NewEvent()
ButtonSelectedEvent, EVT_BUTTON_SELECTED = NewEvent()
NumEditChangedEvent, EVT_NUM_EDIT_CHANGED = NewEvent()
CellSelectedEvent, EVT_CELL_SELECTED = NewEvent()
CellChangedEvent, EVT_CELL_CHANGED = NewEvent()
TextModChangedEvent, EVT_TEXT_MOD_CHANGED = NewEvent()
TextStateChangedEvent, EVT_TEXT_STATE_CHANGED = NewEvent()
LeftClickEvent, EVT_LEFT_CLICK = NewEvent()
RightClickEvent, EVT_RIGHT_CLICK = NewEvent()
MiddleClickEvent, EVT_MIDDLE_CLICK = NewEvent()

custom_map = {
    SIGNAL_BUTTONSELECTED: ButtonSelectedEvent,
    SIGNAL_NUMEDITCHANGED: NumEditChangedEvent,
    SIGNAL_CELLSELECTED: CellSelectedEvent,
    SIGNAL_CELLCHANGED: CellChangedEvent,
    SIGNAL_TEXTMODCHANGED: TextModChangedEvent,
    SIGNAL_TEXTSTATECHANGED: TextStateChangedEvent,
    SIGNAL_QUERYCLOSE: QueryCloseEvent,
    SIGNAL_CLOSING: ClosingEvent,
    SIGNAL_SHOWN: ShownEvent,
    SIGNAL_HIDDEN: HiddenEvent,
    SIGNAL_LEFTCLICK: LeftClickEvent,
    SIGNAL_RIGHTCLICK: RightClickEvent,
    SIGNAL_MIDDLECLICK: MiddleClickEvent,
}

event_map = {
    SIGNAL_ACTIVATED: wx.EVT_MENU,
    SIGNAL_CLICKED: wx.EVT_BUTTON,
    SIGNAL_TOGGLED: wx.EVT_CHECKBOX,
    SIGNAL_SELECTED: wx.EVT_COMBOBOX,
    SIGNAL_BUTTONSELECTED: EVT_BUTTON_SELECTED,
    SIGNAL_FOCUS_IN: wx.EVT_SET_FOCUS,
    SIGNAL_FOCUS_OUT: wx.EVT_KILL_FOCUS,
    SIGNAL_LEFTCLICK: EVT_LEFT_CLICK,
    SIGNAL_RIGHTCLICK: EVT_RIGHT_CLICK,
    SIGNAL_MIDDLECLICK: EVT_MIDDLE_CLICK,
    SIGNAL_LEFTDBLCLICK: wx.EVT_LEFT_DCLICK,
    SIGNAL_RIGHTDBLCLICK: wx.EVT_RIGHT_DCLICK,
    SIGNAL_MIDDLEDBLCLICK: wx.EVT_MIDDLE_DCLICK,
    SIGNAL_LISTBOXSELECTED: wx.EVT_LISTBOX,
    SIGNAL_LISTVIEWSELECTED: wx.dataview.EVT_TREELIST_SELECTION_CHANGED,
    SIGNAL_CELLSELECTED: EVT_CELL_SELECTED,
    SIGNAL_TABLECELLSELECTED: wx.grid.EVT_GRID_SELECT_CELL,
    SIGNAL_TABLECELLEDITING: wx.grid.EVT_GRID_EDITOR_SHOWN,
    SIGNAL_TABLECELLEDITDONE: wx.grid.EVT_GRID_CELL_CHANGING,
    SIGNAL_CELLCHANGED: EVT_CELL_CHANGED,
    SIGNAL_TABLECHANGED: wx.grid.EVT_GRID_CELL_CHANGED,
    SIGNAL_TEXTCHANGED: wx.EVT_TEXT,
    SIGNAL_TEXTMODCHANGED: EVT_TEXT_MOD_CHANGED,
    SIGNAL_TEXTSTATECHANGED: EVT_TEXT_STATE_CHANGED,
    SIGNAL_EDITCHANGED: wx.EVT_TEXT,
    SIGNAL_NUMEDITCHANGED: EVT_NUM_EDIT_CHANGED,
    SIGNAL_VALUECHANGED: wx.EVT_SPINCTRL,
    SIGNAL_ENTER: wx.EVT_TEXT_ENTER,
    SIGNAL_TABSELECTED: wx.EVT_NOTEBOOK_PAGE_CHANGED,
    SIGNAL_QUERYCLOSE: EVT_QUERYCLOSE,
    SIGNAL_CLOSING: EVT_CLOSING,
    SIGNAL_SIZEEVENT: wx.EVT_SIZE,
    SIGNAL_SHOWEVENT: wx.EVT_SHOW,
    SIGNAL_MINIMIZEEVENT: wx.EVT_ICONIZE,
    SIGNAL_SHOWN: EVT_SHOWN,
    SIGNAL_HIDDEN: EVT_HIDDEN,
    SIGNAL_CLOSEEVENT: wx.EVT_CLOSE,
    SIGNAL_BEFOREQUIT: wx.EVT_WINDOW_DESTROY
}


# This is provided to support the custom API; it must be called before
# any widget classes are declared

# Note: args are ignored for this toolkit (unlike pyside2, qt5)

def add_custom_signal(signame, *args):
    # Importing all from defs above will have retrieved the custom signal by name
    sig = getattr(sys.modules[__name__], signame)
    event, binder = NewEvent()
    custom_map[sig] = event
    event_map[sig] = binder


# 'Wrapper' functions for certain events to repackage parameters

def wx_plain_wrapper(self, target):
    def wrapper(event):
        target()
    return wrapper


def wx_toggled_wrapper(self, target):
    def wrapper(event):
        target(self.GetValue())
    return wrapper


def wx_selected_wrapper(self, target):
    def wrapper(event):
        target(self.GetSelection())
    return wrapper


def wx_listviewselected_wrapper(self, target):
    def wrapper(event):
        # wx doesn't use the "list of lists" model so we have to use our current item method
        target(self.current_item())
    return wrapper


def wx_listboxselected_wrapper(self, target):
    def wrapper(event):
        target(self.GetSelection())
    return wrapper


def wx_cellselected_wrapper(self, target):
    def wrapper(event):
        target(event.CurrentRow, event.CurrentCol, event.PreviousRow, event.PreviousCol)
    return wrapper


def wx_cellchanged_wrapper(self, target):
    def wrapper(event):
        target(event.Row, event.Col)
    return wrapper


def wx_textmodchanged_wrapper(self, target):
    def wrapper(event):
        target(self.IsModified())
    return wrapper


def wx_editchanged_wrapper(self, target):
    def wrapper(event):
        target(self.GetValue())
    return wrapper


def wx_numeditchanged_wrapper(self, target):
    def wrapper(event):
        target(int(self.GetValue()))
    return wrapper


def wx_tabselected_wrapper(self, target):
    def wrapper(event):
        target(event.GetSelection())
    return wrapper


wrapper_map = {
    SIGNAL_ACTIVATED: wx_plain_wrapper,
    SIGNAL_CLICKED: wx_plain_wrapper,
    SIGNAL_TOGGLED: wx_toggled_wrapper,
    SIGNAL_SELECTED: wx_selected_wrapper,
    SIGNAL_FOCUS_IN: wx_plain_wrapper,
    SIGNAL_FOCUS_OUT: wx_plain_wrapper,
    SIGNAL_LEFTDBLCLICK: wx_plain_wrapper,
    SIGNAL_RIGHTDBLCLICK: wx_plain_wrapper,
    SIGNAL_MIDDLEDBLCLICK: wx_plain_wrapper,
    SIGNAL_LISTVIEWSELECTED: wx_listviewselected_wrapper,
    SIGNAL_LISTBOXSELECTED: wx_listboxselected_wrapper,
    SIGNAL_TEXTCHANGED: wx_plain_wrapper,
    SIGNAL_EDITCHANGED: wx_editchanged_wrapper,
    SIGNAL_NUMEDITCHANGED: wx_numeditchanged_wrapper,
    SIGNAL_VALUECHANGED: wx_editchanged_wrapper,
    SIGNAL_ENTER: wx_plain_wrapper,
    SIGNAL_TABSELECTED: wx_tabselected_wrapper,
}

no_wrapper_events = set([
    SIGNAL_TABLECELLSELECTED,
    SIGNAL_TABLECELLEDITING,
    SIGNAL_TABLECELLEDITDONE,
    SIGNAL_TABLECHANGED,
    SIGNAL_SIZEEVENT,
    SIGNAL_SHOWEVENT,
    SIGNAL_MINIMIZEEVENT,
    SIGNAL_CLOSEEVENT,
    SIGNAL_FINISHED,  # this is a special case, the handler just takes a boolean
])


def args_wrapper(target):
    def wrapper(event):
        target(*event.args)
    return wrapper


# This event is for manual do_notify, since we can't instantiate wx.Event itself

NotifyEvent, _EVT_NOTIFY = NewEvent()


class PWxSignal(object):
    """Mixin class to abstract notification functionality in wxWidgets.
    """
    
    _forward = None
    
    def wx_register_event(self, event, handler):
        # Do the following instead of self.Bind(event, handler)
        # so that multiple handlers can receive a single event
        eventManager.Register(handler, event, self)
    
    def wrap_target(self, signal, target):
        target = super(PWxSignal, self).wrap_target(signal, target)
        if signal in wrapper_map:
            # These events come from the toolkit internals and aren't used with do_notify, so the parameters
            # the event handler expects need to be extracted from the wx Event object by the wrapper
            return wrapper_map[signal](self, target)
        if signal in no_wrapper_events:
            # These events are handled internally by a widget, so the handler takes the event parameter
            # directly; these events cannot be used with do_notify
            return target
        # These events are used with do_notify in user code and need the do_notify args
        return args_wrapper(target)
    
    def connect_target(self, signal, target):
        if signal in event_map:
            event = event_map[signal]
            self.wx_register_event(event, target)
    
    def do_notify(self, signal, *args):
        if signal in custom_map:
            event = custom_map[signal](args=args)
        else:
            assert signal in event_map
            # This is a kludge because wx doesn't allow you to instantiate wx.Event
            # and then set a type, and it doesn't give an easy way to instantiate
            # all the necessary specific event types
            event = NotifyEvent(args=args)
            event.SetEventType(event_map[signal]._getEvtType())
        if event is not None:
            if self._forward:
                forward_event = self._forward
                event.SetId(forward_event.GetId())
                event.SetEventObject(forward_event.GetEventObject())
                self._forward = None
            else:
                if hasattr(self, 'id'):
                    event.SetId(self.id)
                elif hasattr(self, '_id'):
                    event.SetId(self._id)
                elif hasattr(self, 'GetId'):
                    event.SetId(self.GetId())
                event.SetEventObject(self)
            self.AddPendingEvent(event)
    
    def forward_event(self, signal, event, *args):
        # Ugly hack to allow data to be transferred to custom events
        # that are "translations" of built-in ones
        self._forward = event
        self.do_notify(signal, *args)


def scaled_bitmap(image, factor):
    if factor is not None:
        image = image.Scale(image.GetWidth() * factor,
                            image.GetHeight() * factor)
    return wx.Bitmap(image)


class PWxActionMixin(object):
    
    def load_icon_from_data(self, data):
        import io
        stream = io.BytesIO(data)
        return wx.Image(stream)
    
    def load_icon_from_file(self, filename):
        return wx.Image(filename)
    
    def get_accel_str(self, key):
        return accel_map.get(key)


default_font_size = wx.DEFAULT
font_weights = [wx.FONTWEIGHT_NORMAL, wx.FONTWEIGHT_BOLD]
font_styles = [wx.FONTSTYLE_NORMAL, wx.FONTSTYLE_ITALIC]


def wx_font_object(font_name, font_size=None, bold=None, italic=None):
    font_family = wx.FONTFAMILY_DEFAULT
    for family, names in font_families.items():
        if font_name in names:
            font_family = family
            break
    font_size = font_size or default_font_size
    font_style = font_styles[int(bool(italic))]
    font_weight = font_weights[int(bool(bold))]
    font = wx.Font(font_size, font_family, font_style, font_weight)
    font.SetFaceName(font_name)
    return font


class PWxWidget(PWxSignal):
    """Mixin class to provide basic wx widget methods.
    """
    
    fn_enable_get = 'IsEnabled'
    fn_enable_set = 'Enable'
    
    _depth_test_str = string.ascii_letters + string.digits
    _depth_scale_factor = 1.3
    
    def update_widget(self):
        self.Refresh()
    
    def preferred_width(self):
        return self.GetSize().GetWidth()
    
    def preferred_height(self):
        return self.GetSize().GetHeight()
    
    get_width = preferred_width
    
    get_height = preferred_height
    
    def set_size(self, width, height):
        self.SetSize(width, height)
    
    def get_left(self):
        return self.GetPosition().x
    
    def get_top(self):
        return self.GetPosition().y
    
    def set_position(self, left, top):
        self.SetPosition(wx.Point(left, top))
    
    def set_min_size(self, width, height):
        self.SetMinSize((width, height))
    
    def _mapped_color(self, color):
        if isinstance(color, wx.Colour):
            return color
        return color_map[color]
    
    def set_foreground_color(self, color):
        self.SetForegroundColour(self._mapped_color(color))
    
    def set_background_color(self, color):
        self.SetBackgroundColour(self._mapped_color(color))
    
    def get_font_name(self):
        return self.GetFont().GetFaceName()
    
    def get_font_size(self):
        return self.GetFont().GetPointSize()
    
    def get_font_bold(self):
        return (self.GetFont().GetWeight() == font_weights[1])
    
    def get_font_italic(self):
        return (self.GetFont().GetStyle() == font_styles[1])
    
    def set_font_object(self, font_name, font_size, bold, italic):
        font = wx_font_object(font_name, font_size, bold, italic)
        if hasattr(self, '_depth_w'):
            # Hack to make list view column auto-sizing work correctly
            # on font change
            old_extent = self.GetTextExtent(self._depth_test_str)[0]
            new_extent = self.GetFullTextExtent(self._depth_test_str, font)[0]
            self._depth_w = int(
                self._depth_w * self._depth_scale_factor * new_extent /
                old_extent)
        self.SetFont(font)
        if hasattr(self, 'SetDefaultCellFont'):
            self.SetDefaultCellFont(font)  # takes care of the table widget
        if hasattr(self, 'AutoSizeCols'):
            self.AutoSizeCols()  # this takes care of the list view widget
    
    def set_focus(self):
        self.SetFocus()


# Ugly hack to fix metaclass conflict for sequence widgets

WxMeta = type(wx.Object)


class PWxSequenceMeta(WxMeta, ABCMeta):
    
    def __init__(cls, name, bases, attrs):
        WxMeta.__init__(cls, name, bases, attrs)
        ABCMeta.__init__(cls, name, bases, attrs)


class PWxSequenceWidget(PWxWidget, metaclass=PWxSequenceMeta):
    
    # Ugly hack because the wx event manager needs widget instances
    # to be hashable; note that this is breaking the Python guideline
    # that mutable objects should not be hashable, but we have no
    # choice, and the hash value itself won't change if the widget
    # is mutated
    
    def __hash__(self):
        return hash(id(self))


default_map = {
    wx.ID_CANCEL: wx.CANCEL_DEFAULT,
    wx.ID_NO: wx.NO_DEFAULT,
    wx.ID_OK: wx.OK_DEFAULT,
    wx.ID_YES: wx.YES_DEFAULT,
}


class PWxMessageBox(PMessageBoxBase):
    """Customized wxWidgets message box.
    """
    
    question_map = {
        ANSWER_YES: wx.ID_YES,
        ANSWER_NO: wx.ID_NO,
        ANSWER_CANCEL: wx.ID_CANCEL,
        ANSWER_OK: wx.ID_OK
    }
    
    def message_box(self, type, caption, text, default,
                    button1, button2=None, button3=None):
        
        style = icon_map[type]
        buttons = [button1, button2, button3]
        if (wx.ID_YES in buttons) or (wx.ID_NO in buttons):
            style = style | wx.YES_NO
        if wx.ID_OK in buttons:
            style = style | wx.OK
        if wx.ID_CANCEL in buttons:
            style = style | wx.CANCEL
        default_style = default_map.get(default)
        if default_style:
            style |= default_style
        dlg = wx.MessageDialog(self.parent, text, caption, style)
        # TODO: Hack to fix strange button ordering for Yes/No/Cancel
        #b_cancel = dlg.FindWindowById(wx.ID_CANCEL)
        #b_no = dlg.FindWindowById(wx.ID_NO)
        #print(b_cancel, b_no)
        #if (b_cancel is not None) and (b_no is not None):
        #    b_yes = dlg.FindWindowById(wx.ID_YES)
        #    if b_yes:
        #        b_no.MoveAfterInTabOrder(b_yes)
        #    b_cancel.MoveAfterInTabOrder(b_no)
        result = dlg.ShowModal()
        dlg.Destroy()
        return result
    
    value_type_map = {
        str: wx.TextEntryDialog,
        int: (lambda parent, prompt, caption: wx.NumberEntryDialog(parent, "", prompt, caption,
                                                                   0, (- 2**31), (2**31 - 1))),
    }
    
    def enter_value(self, value_type, caption, prompt):
        f = self.value_type_map.get(value_type)
        if not f:
            raise ValueError("Unsupported value type: {}".format(value_type))
        dlg = f(self.parent, prompt, caption)
        code = dlg.ShowModal()
        result = dlg.GetValue() if code == wx.ID_OK else None
        dlg.Destroy()
        return result


class PWxFileDialog(PFileDialogBase):
    
    def choose_directory(self, curdir):
        dlg = wx.DirDialog(self.parent, defaultPath=curdir)
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            retval = dlg.GetPath()
        else:
            retval = ""
        dlg.Destroy()
        return retval
    
    def wx_filedialog(self, msg, path, filter, style):
        if filter == "":
            filter = "*"
        dlg = wx.FileDialog(self.parent, msg, path, "", filter, style)
        result = dlg.ShowModal()
        if result == wx.ID_OK:
            retval = dlg.GetPath()
        else:
            retval = ""
        dlg.Destroy()
        return retval
    
    def _translate_filter_item(self, caption, ext_str):
        return "{0} ({1})|{1}".format(caption, ext_str)
    
    ext_sep = ";"
    filter_sep = "|"
    
    def _open_filename(self, path, filter, selected_filter):
        # TODO: how to use selected_filter?
        return self.wx_filedialog("Select file to open", path, filter,
                                  wx.FD_OPEN)
    
    def _save_filename(self, path, filter, selected_filter):
        # TODO: how to use selected_filter?
        return self.wx_filedialog("Select file to save", path, filter,
                                  wx.FD_SAVE)


def wx_icon_from_file(filename):
    # Kludgy but gets the job done on all platforms
    return wx.Icon(filename)


class PWxAboutDialog(PAboutDialogBase):
    
    attr_map = {
        'name': "SetName",
        'version': "SetVersion",
        'description': "SetDescription",
        'copyright': "SetCopyright",
        'license': "SetLicense",
        'developer': "AddDeveloper",
        'website': "SetWebSite",
    }
    
    info = None
    
    def display_func(self, parent, caption, body):
        # In case Wx-required about box fields are not present
        parent.app.message_box.info(caption, body)
    
    def display(self):
        data = self.parent.app.about_data
        if any(attr not in data for attr in self.attr_map):
            PAboutDialogBase.display(self)
        else:
            if self.info is None:
                self.info = info = wx.adv.AboutDialogInfo()
                for attr, fname in self.attr_map.items():
                    method = getattr(info, fname)
                    method(data[attr])
                iconfile = self.parent.app.main_iconfile
                if iconfile:
                    info.SetIcon(wx_icon_from_file(iconfile))
            else:
                info = self.info
            wx.adv.AboutBox(info, self.parent)


class PWxProgressDialog(PProgressDialogBase):
    
    def create_dialog(self, title, msg, max_num, stop_label, main_window):
        style = wx.PD_APP_MODAL | wx.PD_AUTO_HIDE
        if stop_label is not None:
            style |= wx.PD_CAN_ABORT
        dialog = wx.ProgressDialog(title, msg, max_num, main_window,
                                   style=style)
        if stop_label is not None:
            button = dialog.GetChildren()[-1]  # hack to get the cancel button
            button.SetLabel(stop_label)
        return dialog
    
    def show_dialog(self):
        self.dialog.Show()
    
    def was_canceled(self):
        return self.dialog.WasCancelled()
    
    def update_progress(self, num):
        self.dialog.Update(num)
    
    def complete(self):
        self.dialog.Destroy()
    
    def close_dialog(self):
        self.dialog.Destroy()


class PTopWindow(PWxWidget, wx.Frame, PTopWindowBase):
    """Customized wxWidgets top window class.
    """
    
    def __init__(self, manager):
        wx.Frame.__init__(self, None)
        PTopWindowBase.__init__(self, manager)
        
        # 'automagic' connections
        self.setup_notify(SIGNAL_CLOSEEVENT, self.OnCloseWindow)
        self.setup_notify(SIGNAL_SHOWEVENT, self.OnShowEvent)
        self.setup_notify(SIGNAL_MINIMIZEEVENT, self.OnIconizeEvent)
    
    def show_init(self):
        PTopWindowBase.show_init(self)
        self.Show(True)
    
    def set_caption(self, caption):
        self.SetTitle(caption)
    
    def get_desktop_rect(self, primary=True):
        r = wx.Display().GetClientArea()
        return (r.GetX(), r.GetY(), r.GetWidth(), r.GetHeight())
    
    def set_client_area_size(self, width, height):
        self.SetClientSize(width, height)
    
    def size_maximize(self):
        self.Maximize(True)
    
    def get_frame_size(self):
        pass  # centering is done differently in wx, see next method
    
    def center(self):
        self.CenterOnScreen()
    
    def do_exit(self):
        self.Close()
    
    def set_iconfile(self, iconfile):
        self.SetIcon(wx_icon_from_file(iconfile))
    
    def get_current_geometry(self):
        left, top = self.GetPosition()
        width, height = self.GetSizeTuple()
        return left, top, width, height
    
    def SignalShown(self, event, shown):
        # 'automagic' method for SIGNAL_SHOWEVENT
        if shown:
            self.forward_event(SIGNAL_SHOWN, event)
        else:
            self.forward_event(SIGNAL_HIDDEN, event)
    
    def OnShowEvent(self, event):
        # 'automagic' method for SIGNAL_SHOWEVENT
        self.SignalShown(event, event.IsShown())
    
    def OnIconizeEvent(self, event):
        # 'automagic' method for SIGNAL_MINIMIZEEVENT
        self.SignalShown(event, not event.IsIconized())
    
    def OnCloseWindow(self, event):
        # 'automagic' method for SIGNAL_CLOSEEVENT
        self.forward_event(SIGNAL_QUERYCLOSE, event)  # a handler here can adjust state to be checked below
        self.ProcessPendingEvents()  # force the handler to run before we check
        
        if event.CanVeto() and not self.can_close():
            event.Veto()
        else:
            # Send the closing signal
            self.forward_event(SIGNAL_CLOSING, event)
            # Allow signal handlers to process before destroying
            self.DestroyLater()


class PApplication(PWxSignal, wx.App, PApplicationBase):
    """Customized wxWidgets application class.
    """
    
    about_dialog_class = PWxAboutDialog
    message_box_class = PWxMessageBox
    file_dialog_class = PWxFileDialog
    progress_dialog_class = PWxProgressDialog
    
    def __init__(self, arglist=[]):
        PApplicationBase.__init__(self, arglist)
        wx.App.__init__(self, arglist)
    
    def setup_notify(self, signal, target):
        # wx.App can't register events, so we fake them as coming from the main window
        self.main_window.setup_notify(signal, target)
    
    def do_notify(self, signal, *args):
        # wx.App can't send events, so we fake them as coming from the main window
        self.main_window.do_notify(signal, *args)
    
    def about_toolkit(self):
        # No built-in "About Wx" dialog
        self.message_box.info("About Wx", "WxPython {}".format(wx.version()))
    
    def OnInit(self):
        # wxWidgets wants you to initialize subwidgets here
        self.do_create()
        
        # required return value
        return True
    
    def OnExit(self):
        # 'automagic' method for SIGNAL_BEFOREQUIT
        self.before_quit()
        
        # required return value
        return super(PApplication, self).OnExit()
    
    def event_loop(self):
        self.MainLoop()
    
    def process_events(self):
        self.ProcessPendingEvents()



#!/usr/bin/env python3
"""
UI-SIGNALS.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A demo app that tests all of the available UI signals.
It also demonstrates how to construct a user interface
dynamically based on information at run time, instead
of having everything pre-defined.
"""

import sys
import os
import functools
import time

from plib.stdlib.builtins import first


# Must do this before any other plib.ui imports or action
# (Note that we don't set a specific toolkit here, but apps
# that want to do that must do it in this section)

from plib.ui import custom

custom.add_signals(
    SIGNAL_CUSTOM1=(('SIGNAL_FINISHED', 50), 'custom1'),
    SIGNAL_CUSTOM2=(('SIGNAL_FINISHED', 51), 'custom2'),
    SIGNAL_EMITTED=(('SIGNAL_FINISHED', 60), 'emitted', int),
)

custom.add_widget_prefixes(
    ClickableLabel='label',
    CustomWidget='widget',
)


from plib.ui import __version__
from plib.ui.defs import *
from plib.ui.app import PApplication, PTextLabel, PMouseMixin
from plib.ui.common import set_italic
from plib.ui.dialogs import StringSelectDialog
from plib.ui.manager import PWidgetManager
from plib.ui.output import PTextOutput
from plib.ui.widgets import *


class MouseMixin(PMouseMixin):
    
    @property
    def event_signal(self):
        return self.signals[0]
    
    msg_signames = (
        'SIGNAL_LEFTCLICK',
        'SIGNAL_RIGHTCLICK',
        'SIGNAL_MIDDLECLICK',
        'SIGNAL_LEFTDBLCLICK',
        'SIGNAL_RIGHTDBLCLICK',
        'SIGNAL_MIDDLEDBLCLICK',
    )
    
    def setup_signals(self):
        super(MouseMixin, self).setup_signals()
        for signame in self.msg_signames:
            sig = getattr(sys.modules[__name__], signame)
            self.setup_notify(sig, functools.partial(self.on_mouse_signal, signame))
        self.setup_notify(SIGNAL_LEFTCLICK, self.on_left_click)
    
    def on_mouse_signal(self, signame):
        self.manager.app.output_message("From Page {}: {}".format(self.__class__.__name__, signame))
    
    def on_left_click(self):
        self.do_notify(self.event_signal, *self.event_args)


class ClickableLabel(MouseMixin, PTextLabel):
    
    signals = (
        SIGNAL_CLICKED,
    )
    
    event_args = ()


class CustomWidget(MouseMixin, PTextLabel):
    
    signals = (
        SIGNAL_EMITTED,
    )
    
    @property
    def event_args(self,
                   sto=[0]):
        
        num = sto[0]
        sto[0] += 1
        return (num,)


def clickable_label(name, caption, geometry=None, font=None):
    args = (caption,)
    kwargs = dict(
        geometry=geometry,
        font=font,
    )
    return ('__main__', 'ClickableLabel', name, args, kwargs)


def custom_widget(name, caption, geometry=None, font=None):
    args = (caption,)
    kwargs = dict(
        geometry=geometry,
        font=font,
    )
    return ('__main__', 'CustomWidget', name, args, kwargs)


def display_sub_objects(mgr, mgr_name='app'):
    print("Showing sub-objects for {}".format(mgr_name))
    for widget in mgr.all_widgets:
        name = first(attr for attr in dir(mgr) if getattr(mgr, attr) is widget)
        print("Widget: {}".format(name))
        assert getattr(mgr, name) is widget
    for manager in mgr.sub_managers:
        name = manager.attrname
        print("Manager: {}".format(name))
        assert getattr(mgr, name) is manager


number_strings = ("One", "Two", "Three", "Four", "Five", "Six")


class Page(PWidgetManager):
    
    name = 'manager'
    
    @property
    def main_widget(self):
        # We implement this as a property because we can't wait until before_create
        # to define the widget specs (as is done in the app below); the widget_from_spec
        # function gets called with this class and evaluates main_widget before any
        # widget creation starts. This will be true for any widget manager that is
        # used in the app's main_widget spec.
        return tab(ALIGN_JUST, LAYOUT_VERTICAL, [
            box(ALIGN_TOP, LAYOUT_VERTICAL, [
                clickable_label('labeltest', "Clickable Label Test"),
            ]),
            listbox('numbers', number_strings, font=self.app.font),
            box(ALIGN_BOTTOM, LAYOUT_HORIZONTAL, [
                button('custom1', "Custom Signal 1", font=self.app.font),
                button('custom2', "Custom Signal 2", font=self.app.font),
                padding(),
                button('input_str', "Input String", font=self.app.font),
                button('input_int', "Input Integer", font=self.app.font),
                padding(),
                custom_widget('custom3', "Custom Widget", font=self.app.font),
            ]),
        ])
    
    def setup_signals(self):
        # Normally should not need to override this method, we do it here only to show timing of method calls
        PWidgetManager.setup_signals(self)
        print("Page setup_signals")
        # TODO: can we automate this connection so the page knows when it is shown?
        self.app.panels.setup_notify(SIGNAL_TABSELECTED, self.on_page_selected)
        self.target_index = len(self.app.panels) - 1
    
    def after_create(self):
        display_sub_objects(self, self.attrname)
        print("Page after_create")
    
    showing = False  # since this page isn't displayed initially
    
    def on_page_selected(self, index):
        if index == self.target_index:
            self.populate_data()
    
    def populate_data(self):
        # Now the app's output_message function is set up
        self.app.output_message("Page populate_data")
    
    def on_numbers_selected(self, index):
        assert index == self.listbox_numbers.current_index()
        assert self.listbox_numbers[index] == self.listbox_numbers.current_text()
        self.app.output_message("From Page: SIGNAL_LISTBOXSELECTED {} {}".format(
            index, self.listbox_numbers.current_text()
        ))
    
    def on_labeltest_clicked(self):
        self.app.output_message("From Page ClickableLabel: SIGNAL_CLICKED")
    
    def on_custom1_clicked(self):
        print("custom1_clicked")
        self.app.do_notify(SIGNAL_CUSTOM1)
    
    def on_custom2_clicked(self):
        print("custom2_clicked")
        self.app.do_notify(SIGNAL_CUSTOM2)
    
    def on_input_str_clicked(self):
        value = self.app.message_box.enter_string("Enter String", "String value:")
        if value is not None:
            self.listbox_numbers.append(value)
    
    def on_input_int_clicked(self):
        value = self.app.message_box.enter_int("Enter Integer", "Integer value:")
        if value is not None:
            self.listbox_numbers.append(str(value))
    
    def on_custom3_emitted(self, num):
        self.app.output_message("From Page CustomWidget: SIGNAL_EMITTED {}".format(num))


PAGE_TABS = 0
PAGE_BUTTONS = 1
PAGE_COMBO = 2
PAGE_LIST = 3
PAGE_RADIO = 4
PAGE_DIALOG = 5


class PageSelectDialog(StringSelectDialog):
    
    @property
    def dialog_client(self):
        # This illustrates doing surgery on a widget spec (note that we can't do this in
        # before_create, even though this dialog is only constructed dynamically after the
        # app starts, because StringSelectDialog defines dialog_client as a property)
        client = super(PageSelectDialog, self).dialog_client
        client[4]['contents'][0][4]['contents'][0][4]['contents'].insert(0,
            label('message', "Current page: {}".format(self.manager.label_current_page.caption),
                  font=set_italic(self.app.font))
        )
        return client


class DialogButton(PWidgetManager):
    
    @property
    def font(self):
        f = self.app.font
        assert (not f) or (len(f) == 1)
        font_name = f[0] if f else None
        return (font_name, None, True)
    
    @property
    def main_widget(self):
        # We do this as a property for the same reason as the Page class above
        # (see comment there for further discussion)
        return panel(ALIGN_TOP, LAYOUT_HORIZONTAL, [
            label('current_page', "", font=self.font),
            button('change_page', "Change...", font=self.app.font),
        ])
    
    titles = None
    
    def setup_signals(self):
        # Normally should not need to override this method, we do it here only to show timing of method calls
        PWidgetManager.setup_signals(self)
        print("DialogButton setup_signals")
    
    def after_create(self):
        display_sub_objects(self, self.attrname)
        print("DialogButton after_create")
        self.label_current_page.caption = self.titles[0]
    
    def current_index(self):
        return self.titles.index(self.label_current_page.caption)
    
    def __getitem__(self, index):
        return self.titles[index]
    
    def page_change_callback(self, title):
        self.label_current_page.caption = title
        # TODO: can we make this signal connection automatic?
        self.manager.on_panels_selector_selected(self.current_index())
        self.manager.pagewidget_panels.set_current_index(self.current_index())
    
    def on_change_page_clicked(self):
        PageSelectDialog(self, "Select Page", self.titles, "Select the page to switch to:",
                         self.label_current_page.caption, self.page_change_callback,
                         font=self.app.font).run()


def dialogbutton(name, titles):
    DialogButton.name = name
    DialogButton.titles = titles
    return DialogButton


def horiz(group):
    def horiz_group(name, titles, font=None):
        return group(name, ALIGN_TOP, LAYOUT_HORIZONTAL, titles, starting_index=0, font=font)
    return horiz_group


def vert(group):
    def vert_group(name, titles, font=None):
        return group(name, ALIGN_LEFT, LAYOUT_VERTICAL, titles, starting_index=0, font=font)
    return vert_group


def listbox_panel(name, titles, font=None):
    return panel(ALIGN_JUST, LAYOUT_VERTICAL, [
        listbox(name, titles, font=font),
    ], name=name)


def panels(name, contents, app):
    sel = UISignalTester.page_select
    sel_layout = UISignalTester.select_layout
    if sel == PAGE_TABS:
        if sel_layout:
            print("Ignoring selector layout option, not usable with tab widget")
        return tabwidget(name, contents, font=app.font)
    
    selector_type, selector, default_panel_layout, selector_kwargs = (
        ('buttongroup', buttongroup, LAYOUT_VERTICAL, dict(font=app.font)) if sel == PAGE_BUTTONS else
        ('combo', combo, LAYOUT_VERTICAL, dict(font=app.font)) if sel == PAGE_COMBO else
        ('listbox', listbox_panel, LAYOUT_HORIZONTAL, dict(font=app.font)) if sel == PAGE_LIST else
        ('radiogroup', radiogroup, LAYOUT_VERTICAL, dict(font=app.font)) if sel == PAGE_RADIO else
        ('dialogbutton', dialogbutton, LAYOUT_VERTICAL, {}) if sel == PAGE_DIALOG else
        (None, None, None, None)
    )
    if (sel_layout == LAYOUT_VERTICAL) and (selector_type == 'combo'):
        raise RuntimeError("Cannot use combo box selector with vertical layout")
    if (sel_layout == LAYOUT_VERTICAL) and (selector_type == 'dialogbutton'):
        raise RuntimeError("Cannot use dialog button selector with vertical layout")
    if (sel_layout == LAYOUT_HORIZONTAL) and (selector_type == 'listbox'):
        raise RuntimeError("Cannot use list box selector with horizontal layout")
    assert selector is not None
    
    oriented, panel_layout = (
        (horiz, LAYOUT_VERTICAL) if sel_layout == LAYOUT_HORIZONTAL else
        (vert, LAYOUT_HORIZONTAL) if sel_layout == LAYOUT_VERTICAL else
        (None, None)
    )
    if panel_layout is None:
        panel_layout = default_panel_layout
    if selector_type in ('buttongroup', 'radiogroup'):
        if oriented is None:
            oriented = (
                horiz if panel_layout == LAYOUT_VERTICAL else
                vert if panel_layout == LAYOUT_HORIZONTAL else
                None
            )
            assert oriented is not None
        selector = oriented(selector)
    
    UISignalTester.selector_type = selector_type
    selector_name = '{}_selector'.format(name)
    frame_name = '{}_frame'.format(name)
    titles = [item[0] for item in contents]
    pages = [item[1] for item in contents]
    link_to = (
        '{}_{}'.format(selector_type, selector_name)
        if selector_type != 'dialogbutton' else
        None  # TODO: can we set up a selected signal for this case?
    )
    
    return panel(ALIGN_JUST, panel_layout, [
        selector(selector_name, titles, **selector_kwargs),
        pagewidget(name, pages, link_to=link_to),
    ], name=frame_name)


LONG_DESC = "List view item with long description to test horizontal scrolling behavior"


class UISignalTester(PApplication):
    
    signals = (
        SIGNAL_CUSTOM1,
        SIGNAL_CUSTOM2,
    )
    
    page_select = PAGE_TABS
    select_layout = None
    
    about_data = {
        'name': "UISignalTester",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "UI Signal Test Demo",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net"
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    main_title = "UI Signal Tester"
    
    main_size = SIZE_CLIENTWRAP
    main_placement = MOVE_CENTER
    
    font_name = None
    
    @property
    def font(self):
        return (self.font_name,) if self.font_name else None
    
    def before_create(self):
        # We can't call output_message here because widgets aren't available yet
        print("before_create")
        
        # We do this here so the panels function will see the options
        # filled in by option parsing. We could also do it as a property,
        # as is done in the Page and DialogButton classes above, but we
        # want to illustrate that it *can* be done in before_create for
        # the app. It could also be done in before_create for a widget
        # manager that is not referenced in the app's main_widget spec but
        # is only instantiated dynamically after the app starts.
        self.main_widget = frame(ALIGN_JUST, LAYOUT_HORIZONTAL, [
            panel(ALIGN_JUST, LAYOUT_VERTICAL, [
                panels('panels', [
                    ('Dialog', tab(ALIGN_JUST, LAYOUT_VERTICAL, [
                        button('action', "Test Clicked", font=self.font),
                        checkbox('option', "Test Toggled", font=self.font),
                        box(ALIGN_TOP, LAYOUT_HORIZONTAL, [
                            button('progress', "Test Progress", font=self.font),
                            checkbox('progress_abort', "Include progress abort button", font=self.font),
                            padding(),
                        ]),
                        sorted_combo('selection', (numstr.lower() for numstr in number_strings), font=self.font),
                        edit('text', font=self.font),
                        num_edit('num', font=self.font),
                        spin_edit('spinner', font=self.font),
                        sorted_listbox('list', ("Item {}".format(numstr) for numstr in number_strings), font=self.font),
                        buttongroup('colgrid', ALIGN_BOTTOM, (LAYOUT_COLGRID, 3),
                                    ("Button {}".format(numstr) for numstr in number_strings), font=self.font),
                        radiogroup('rowgrid', ALIGN_BOTTOM, (LAYOUT_ROWGRID, 3),
                                   ("Radio Button {}".format(numstr) for numstr in number_strings), font=self.font),
                    ])),
                    ('Memo', memo('notes', font=self.font)),
                    ('Tree', treeview('tree', [("Title", WIDTH_CONTENTS), ("Description", WIDTH_STRETCH)], (
                        (("Item {}".format(numstr), "Tree item {}".format(numstr.lower())), tuple(
                            (("Sub-item {} {}".format(numstr, substr), "Tree sub-item {} {}".format(numstr, substr)), ())
                            for substr in number_strings[:3]
                        ))
                        for numstr in number_strings
                    ), font=self.font, auto_expand=True)),
                    ('List', sorted_listview('items', [("Title", WIDTH_CONTENTS), ("Description", WIDTH_STRETCH)], (
                        ("Item {}".format(numstr), "List item {}".format(numstr.lower()))
                        for numstr in number_strings
                    ), font=self.font)),
                    ('Table', table('cells', [("Title", WIDTH_CONTENTS), ("Description", WIDTH_STRETCH)], (
                        ("Row {}".format(numstr), "Table row {}".format(numstr.lower()))
                        for numstr in number_strings
                    ), font=self.font)),
                    ('Scroller', scrolling(listbox('scroll_list', (
                        "{} {} {} - {}".format(a, b, c, LONG_DESC)
                        for a in number_strings for b in number_strings for c in number_strings
                    ), font=self.font))),
                    ('Page', Page),
                ], self),
                box(ALIGN_BOTTOM, LAYOUT_HORIZONTAL, [
                    padding(),
                    button('reset', "Reset Selector", font=self.font),
                ]),
            ]),
            text('output', font=self.font),
        ])
    
    def setup_signals(self):
        # Normally should not need to override this method, we do it here only to show timing of method calls
        PApplication.setup_signals(self)
        print("setup_signals")
    
    selector_type = None
    selector = None
    
    selector_sig_names = {
        'buttongroup': "BUTTON",
        'combo': "",
        'listbox': "LISTBOX",
        'radiogroup': "BUTTON",
        'dialogbutton': "",
    }
    
    selector_sig_name = None
    
    @property
    def panels(self):
        return getattr(self, '{}widget_panels'.format('tab' if self.page_select == PAGE_TABS else 'page'))
    
    def after_create(self):
        display_sub_objects(self)
        
        self.outputfile = PTextOutput(self.text_output)
        self.output_message("after_create")
        
        if self.page_select == PAGE_TABS:
            self.panel_title = (lambda index: self.tabwidget_panels[index][0])
        else:
            self.selector = selector = getattr(self, '{}_panels_selector'.format(self.selector_type))
            self.panel_title = (lambda index: selector[index])
            self.selector_sig_name = "SIGNAL_{}SELECTED".format(self.selector_sig_names[self.selector_type])
    
    def populate_data(self):
        self.output_message("populate_data")
    
    def output_message(self, message):
        print(message)
        self.outputfile.write("{}{}".format(message, os.linesep))
        self.outputfile.flush()
    
    def on_reset_clicked(self):
        self.output_message("SIGNAL_CLICKED")
        if self.selector is None:
            self.message_box.info("Information", "No reset available with tab widget.")
        elif isinstance(self.selector, DialogButton):
            self.message_box.info("Information", "No reset available with dialog button.")
        else:
            index = self.selector.current_index()
            titles = self.selector[:]
            self.selector.load_items(titles, starting_index=index)
    
    def focus_in(self, widget_name):
        self.output_message("SIGNAL_FOCUS_IN {}".format(widget_name))
    
    def focus_out(self, widget_name):
        self.output_message("SIGNAL_FOCUS_OUT {}".format(widget_name))
    
    def on_text_focus_in(self):
        self.focus_in('edit_text')
    
    def on_text_focus_out(self):
        self.focus_out('edit_text')
    
    def on_notes_focus_in(self):
        self.focus_in('memo_notes')
    
    def on_notes_focus_out(self):
        self.focus_out('memo_notes')
    
    def on_panels_selector_selected(self, index):
        assert self.selector.current_index() == index
        self.output_message("{} {}".format(self.selector_sig_name, index))
    
    def on_panels_selected(self, index):
        assert self.panels.current_index() == index
        self.output_message("SIGNAL_TABSELECTED {} {}".format(
            index, self.panel_title(index))
        )
    
    def on_action_clicked(self):
        self.output_message("SIGNAL_CLICKED")
    
    def on_option_toggled(self, checked):
        assert checked == self.checkbox_option.checked
        self.output_message("SIGNAL_TOGGLED {}".format(('off', 'on')[checked]))
    
    def on_progress_clicked(self):
        max_num = 20
        stop_label = "Abort Test" if self.checkbox_progress_abort.checked else None
        with self.progress_dialog.updater("Progress Updater Test", "Test Progress", max_num, stop_label) as p:
            for n in range(max_num):
                if p.was_canceled():
                    self.output_message("Progress aborted")
                    p.complete()
                    break
                self.output_message("Progress update: {}".format(n))
                p.update_progress(n)
                time.sleep(0.25)
            else:
                self.output_message("Progress update: {}".format(max_num))
                p.update_progress(max_num)
    
    def on_progress_abort_toggled(self, checked):
        self.output_message("SIGNAL_TOGGLED {}".format(('off', 'on')[checked]))
    
    def on_selection_selected(self, index):
        assert self.combo_selection.current_index() == index
        assert self.combo_selection.current_text() == self.combo_selection[index]
        self.output_message("SIGNAL_SELECTED {} {}".format(
            index, self.combo_selection.current_text()
        ))
    
    def on_text_changed(self, text):
        assert text == self.edit_text.edit_text
        self.output_message("SIGNAL_EDITCHANGED {}".format(text))
    
    def on_text_enter(self):
        self.output_message("SIGNAL_ENTER")
    
    def on_num_value_changed(self, value):
        assert value == self.edit_num.edit_value
        self.output_message("SIGNAL_NUMEDITCHANGED {}".format(value))
    
    def on_num_enter(self):
        self.output_message("SIGNAL_ENTER")
    
    def on_spinner_value_changed(self, value):
        assert value == self.edit_spinner.edit_value
        self.output_message("SIGNAL_VALUECHANGED {}".format(value))
    
    def on_list_selected(self, index):
        assert index == self.listbox_list.current_index()
        assert self.listbox_list[index] == self.listbox_list.current_text()
        self.output_message("SIGNAL_LISTBOXSELECTED {} {}".format(
            index, self.listbox_list.current_text()
        ))
    
    def on_colgrid_selected(self, index):
        assert index == self.buttongroup_colgrid.current_index()
        assert self.buttongroup_colgrid[index] == self.buttongroup_colgrid.current_text()
        self.output_message("SIGNAL_SELECTED {} {}".format(
            index, self.buttongroup_colgrid.current_text()
        ))
    
    def on_rowgrid_selected(self, index):
        assert index == self.radiogroup_rowgrid.current_index()
        assert self.radiogroup_rowgrid[index] == self.radiogroup_rowgrid.current_text()
        self.output_message("SIGNAL_SELECTED {} {}".format(
            index, self.radiogroup_rowgrid.current_text()
        ))
    
    def on_notes_changed(self):
        self.output_message("SIGNAL_TEXTCHANGED {}".format(self.memo_notes.edit_text))
    
    def on_notes_mod_changed(self, changed):
        self.output_message("SIGNAL_TEXTMODCHANGED {} {}".format(self.memo_notes.edit_text, changed))
    
    def on_notes_state_changed(self):
        self.output_message("SIGNAL_TEXTSTATECHANGED memo_notes")
    
    def on_tree_selected(self, cols, children):
        assert (cols, children) == (self.treeview_tree.current_item().cols, self.treeview_tree.current_item().children)
        self.output_message("SIGNAL_LISTVIEWSELECTED {} ({}) {} {} {}".format(
            'treeview', " ".join(str(i) for i in self.treeview_tree.current_indexes()), len(children), *cols
        ))
    
    def on_items_selected(self, cols):
        assert cols == self.listview_items.current_item().cols
        assert cols == self.listview_items[self.listview_items.current_index()]
        self.output_message("SIGNAL_LISTVIEWSELECTED {} {} {} {}".format(
            'listview', self.listview_items.current_index(), *cols
        ))
    
    def on_cells_selected(self, curr_row, curr_col, prev_row, prev_col):
        assert self.table_cells.current_row() == curr_row
        assert self.table_cells.current_col() == curr_col
        assert self.table_cells.current_cell() == self.table_cells[curr_row][curr_col]
        self.output_message("SIGNAL_CELLSELECTED ({}, {}) -> ({}, {}) {}".format(
            str(prev_row), str(prev_col), str(curr_row), str(curr_col), self.table_cells.current_cell())
        )
    
    def on_cells_changed(self, row, col):
        assert self.table_cells.current_row() == row
        assert self.table_cells.current_col() == col
        assert self.table_cells.current_cell() == self.table_cells[row][col]
        self.output_message("SIGNAL_CELLCHANGED {} {} {}".format(
            str(row), str(col), self.table_cells.current_cell())
        )
    
    def on_scroll_list_selected(self, index):
        assert index == self.listbox_scroll_list.current_index()
        assert self.listbox_scroll_list[index] == self.listbox_scroll_list.current_text()
        self.output_message("SIGNAL_LISTBOXSELECTED {} {}".format(
            index, self.listbox_scroll_list.current_text()
        ))
    
    def on_app_custom1(self):
        self.output_message("SIGNAL_CUSTOM1")
    
    def on_app_custom2(self):
        self.output_message("SIGNAL_CUSTOM2")
    
    queryclose_handled = False
    
    def on_app_queryclose(self):
        self.queryclose_handled = True
        self.output_message("SIGNAL_QUERYCLOSE")
    
    def accept_close(self):
        print("SIGNAL_QUERYCLOSE handled:", self.queryclose_handled)
        return self.queryclose_handled
    
    def on_app_closing(self):
        self.output_message("SIGNAL_CLOSING")
    
    def on_app_shown(self):
        self.output_message("SIGNAL_SHOWN")
    
    def on_app_hidden(self):
        self.output_message("SIGNAL_HIDDEN")
    
    def before_quit(self):
        # We can't call output_message here because widgets might not be available
        # any more
        print("before_quit")


if __name__ == "__main__":
    from plib.stdlib.options import parse_options
    optlist = (
        ("-b", "--panel-buttons", {
            'action': "store_const", 'const': PAGE_BUTTONS,
           'dest': "page_select", 'default': PAGE_TABS,
            'help': "Use toggle buttons to select panels"
        }),
        ("-c", "--panel-combo", {
            'action': "store_const", 'const': PAGE_COMBO,
            'dest': "page_select", 'default': PAGE_TABS,
            'help': "Use combo box to select panels"
        }),
        ("-d", "--panel-dialog", {
            'action': "store_const", 'const': PAGE_DIALOG,
            'dest': "page_select", 'default': PAGE_TABS,
            'help': "Use dialog box to select panels"
        }),
        ("-f", "--font_name", {
            'action': "store",
            'dest': "font_name",
            'help': "Font name for widgets"
        }),
        ("-l", "--panel-list", {
            'action': "store_const", 'const': PAGE_LIST,
            'dest': "page_select", 'default': PAGE_TABS,
            'help': "Use list box to select panels"
        }),
        ("-r", "--panel-radio", {
            'action': "store_const", 'const': PAGE_RADIO,
            'dest': "page_select", 'default': PAGE_TABS,
            'help': "Use radio buttons to select panels"
        }),
        ("-t", "--panel-tabs", {
            'action': "store_const", 'const': PAGE_TABS,
            'dest': "page_select", 'default': PAGE_TABS,
            'help': "Use tab widget to select panels"
        }),
        ("-v", "--selector-vertical", {
            'action': "store_const", 'const': LAYOUT_VERTICAL,
            'dest': "select_layout", 'default': None,
            'help': "Orient panel selector vertically"
        }),
        ("-z", "--selector-horizontal", {
            'action': "store_const", 'const': LAYOUT_HORIZONTAL,
            'dest': "select_layout", 'default': None,
            'help': "Orient panel selector horizontally"
        }),
    )
    opts, args = parse_options(optlist)
    # The options object supports a dictionary interface,
    # making it easy to update class fields from it
    for opt, value in opts.items():
        setattr(UISignalTester, opt, value)
    UISignalTester().run()



#!/usr/bin/env python3
"""
Module MOUSE -- UI Mouse Widget Mixin
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import wx

from plib.stdlib.builtins import first

from plib.ui.defs import *
from plib.ui.base.mouse import PMouseBase


class PMouseMixin(PMouseBase):
    
    button_signals = {
        wx.MOUSE_BTN_LEFT: SIGNAL_LEFTCLICK,
        wx.MOUSE_BTN_RIGHT: SIGNAL_RIGHTCLICK,
        wx.MOUSE_BTN_MIDDLE: SIGNAL_MIDDLECLICK,
    }
    
    button_pressed = None
    
    def setup_signals(self):
        self.wx_register_event(wx.EVT_LEFT_DOWN, self.OnButtonDown)
        self.wx_register_event(wx.EVT_RIGHT_DOWN, self.OnButtonDown)
        self.wx_register_event(wx.EVT_MIDDLE_DOWN, self.OnButtonDown)
        self.wx_register_event(wx.EVT_LEFT_UP, self.OnButtonUp)
        self.wx_register_event(wx.EVT_RIGHT_UP, self.OnButtonUp)
        self.wx_register_event(wx.EVT_MIDDLE_UP, self.OnButtonUp)
    
    def OnButtonDown(self, event):
        self.button_pressed = first(b for b in self.button_signals if event.Button(b))
        event.Skip()  # so default processing still occurs
    
    def OnButtonUp(self, event):
        button_released = first(b for b in self.button_signals if event.Button(b))
        if (button_released is not None) and (button_released == self.button_pressed):
            self.forward_event(self.button_signals[self.button_pressed], event)
        self.button_pressed = None
        event.Skip()  # so default processing still occurs



#!/usr/bin/env python3
"""
Module WX.TABWIDGET -- Python wxWidgets Tab Widget
Sub-Package UI.TOOLKITS.QT of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for the page widget.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.pagewidget import PPageWidgetBase

from .app import PWxSequenceWidget


class PPageWidget(PWxSequenceWidget, wx.Simplebook, PPageWidgetBase):
    
    _align = True  # used by panel to determine placement
    _expand = True
    
    def __init__(self, manager, parent, pages=None, link_to=None):
        wx.Simplebook.__init__(self, parent)
        PPageWidgetBase.__init__(self, manager, parent, pages=pages, link_to=link_to)
    
    def page_count(self):
        return self.GetPageCount()
    
    def page_at(self, index):
        return self.GetPage(index)
    
    def add_page(self, index, widget):
        # Use empty labels since wx requires the label as a parameter
        if index == self.__len__():
            self.AddPage(widget, "")
        else:
            self.InsertPage(index, widget, "")
    
    def del_page(self, index):
        self.RemovePage(index)
    
    def current_index(self):
        return self.GetSelection()
    
    def set_current_index(self, index):
        self.SetSelection(index)



#!/usr/bin/env python3
"""
Module SOCKET -- UI Socket Wrapper Objects
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import functools

from plib.ui.defs import *
from plib.ui.app import PSocketNotifier


class PSocketHandler(object):
    
    def __init__(self, transport, sock, readable, handle_read, writable, handle_write, done, end_notify):
        self.transport = transport
        self.sock = sock
        self.closed = False
        
        self.read_notifier = transport.notifier_class(sock, NOTIFY_READ, readable, self._wrap(handle_read, done, end_notify))
        self.write_notifier = transport.notifier_class(sock, NOTIFY_WRITE, writable, self._wrap(handle_write, done, end_notify))
    
    def close(self):
        if self.closed:
            return
        self.read_notifier.set_enabled(False)
        self.read_notifier = None  # no way to actually close the notifier, this is the best we can do
        self.write_notifier.set_enabled(False)
        self.write_notifier = None
        self.transport.remove_handler(self.sock)
        self.closed = True
    
    def _wrap(self, fn, done, end_notify):
        @functools.wraps(fn)
        def _f(sock):
            assert sock is self.sock
            fn()
            if done():
                self.close()
                end_notify()
        return _f


class PSocketTransport(object):
    
    notifier_class = PSocketNotifier
    
    nonblocking = True  # for compatibility with plib.stdlib.net
    
    def __init__(self):
        self.handlers = {}
        self.started = False
    
    def add_handler(self, sock, readable, handle_read, writable, handle_write, done, end_notify):
        self.handlers[sock] = PSocketHandler(self, sock, readable, handle_read, writable, handle_write, done, end_notify)
    
    def run(self):
        # This tells the notifier to start processing socket events
        if self.handlers and not self.started:
            self.notifier_class.start()
            self.started = True
    
    def remove_handler(self, sock):
        del self.handlers[sock]
        if self.started and not self.handlers:
            # This tells the notifier to stop processing socket events
            self.notifier_class.done()
            self.started = False



#!/usr/bin/env python3
"""
Module QT5.GROUP -- Python Qt 5 Button Group Objects
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI objects for the panel widgets.
"""

from PyQt5 import QtWidgets as qt

from plib.ui.defs import *
from plib.ui.base.group import PButtonGroupMixin

from .app import PQtSequenceMeta, qt_font_object
from .form import PPanel


class PButtonGroupBase(PButtonGroupMixin, PPanel,
                       metaclass=PQtSequenceMeta):
    
    event_signals = (
        SIGNAL_BUTTONSELECTED,
    )
    
    button_class = None
    group = None
    
    def __init__(self, manager, parent, align, layout, items=None, value=None, starting_index=None, font=None,
                 style=PANEL_PLAIN, margin=None, spacing=None):
        
        PPanel.__init__(self, manager, parent, align, layout,
                        style=style, margin=margin, spacing=spacing)
        PButtonGroupMixin.__init__(self, items, value, starting_index, font)
    
    def init_group(self):
        self.group = qt.QButtonGroup(self)
        self.group.buttonToggled.connect(self.on_button_toggled)
    
    def create_button(self, item):
        return self.button_class(item, self)
    
    def set_button_font(self, button, font_name=None, font_size=None, bold=None, italic=None):
        button.setFont(qt_font_object(font_name, font_size, bold, italic))
    
    def init_button(self, button):
        self.group.addButton(button)
    
    def group_buttons(self):
        return self.group.buttons()
    
    def del_button(self, button):
        self.group.removeButton(button)
    
    def is_checked(self, button):
        return button.isChecked()
    
    def check_button(self, button, checked=True):
        button.setChecked(checked)
    
    def button_text(self, button):
        return button.text()
    
    def set_button_text(self, button, value):
        button.setText(value)


class PButtonGroup(PButtonGroupBase):
    
    button_class = qt.QPushButton
    
    def init_button(self, button):
        PButtonGroupBase.init_button(self, button)
        button.setCheckable(True)


class PRadioGroup(PButtonGroupBase):
    
    button_class = qt.QRadioButton



#!/usr/bin/env python3
"""
Module EXAMPLES -- Entry points for example programs
Sub-Package UI of Package PLIB3
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains entry points for the example programs
shipped with PLIB3.UI.
"""

from plib.stdlib import postinstall


postinstall.make_entry_points(__name__, 'plib3.ui', 'plib.ui', share_root="examples")



#!/usr/bin/env python3
"""
Module QT5.APP -- Python Qt 5 Application Objects
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI application objects.
"""

from abc import ABCMeta
from itertools import chain

from PyQt5 import QtCore as qtc, QtGui as qtg, QtWidgets as qt

from plib.ui.defs import *
from plib.ui.base.app import (
    PMessageBoxBase, PFileDialogBase, PAboutDialogBase, PProgressDialogBase,
    PTopWindowBase, PApplicationBase
)


align_map = {
    ALIGN_LEFT: qtc.Qt.AlignLeft | qtc.Qt.AlignVCenter,
    ALIGN_CENTER: qtc.Qt.AlignCenter,
    ALIGN_RIGHT: qtc.Qt.AlignRight | qtc.Qt.AlignVCenter
}

color_map = dict(
    (color, qtg.QColor(color.lower()))
    for color in COLORNAMES
)

accel_map = {
    ACTION_FILE_NEW: qtg.QKeySequence.New,
    ACTION_FILE_OPEN: qtg.QKeySequence.Open,
    ACTION_FILE_SAVE: qtg.QKeySequence.Save,
    ACTION_FILE_SAVEAS: qtg.QKeySequence.SaveAs,
    ACTION_FILE_CLOSE: qtg.QKeySequence.Close,
    ACTION_EDIT_UNDO: qtg.QKeySequence.Undo,
    ACTION_EDIT_REDO: qtg.QKeySequence.Redo,
    ACTION_EDIT_CUT: qtg.QKeySequence.Cut,
    ACTION_EDIT_COPY: qtg.QKeySequence.Copy,
    ACTION_EDIT_PASTE: qtg.QKeySequence.Paste,
    ACTION_EDIT_DELETE: qtg.QKeySequence.Delete,
    ACTION_EDIT_SELECTALL: qtg.QKeySequence.SelectAll,
    ACTION_EDIT_SELECTNONE: qtg.QKeySequence.Deselect,
    ACTION_EDIT_OVERWRITE: "Ins",
    #ACTION_EDIT_CLEAR: qtg.QKeySequence.,
    #ACTION_VIEW: qtg.QKeySequence.,
    #ACTION_EDIT: qtg.QKeySequence.,
    #ACTION_OK: qtg.QKeySequence.,
    #ACTION_CANCEL: qtg.QKeySequence,  # FIXME: this is present in PySide
    ACTION_REFRESH: qtg.QKeySequence.Refresh,
    #ACTION_ADD: qtg.QKeySequence.,
    #ACTION_REMOVE: qtg.QKeySequence.,
    #ACTION_APPLY: qtg.QKeySequence.,
    #ACTION_COMMIT: qtg.QKeySequence.,
    #ACTION_ROLLBACK: qtg.QKeySequence.,
    ACTION_PREFS: qtg.QKeySequence.Preferences,
    ACTION_ABOUT: "Alt+B",
    ACTION_ABOUT_TOOLKIT: "Alt+T",
    ACTION_EXIT: qtg.QKeySequence.Quit,
}

message_funcs = {
    MBOX_INFO: qt.QMessageBox.information,
    MBOX_WARN: qt.QMessageBox.warning,
    MBOX_ERROR: qt.QMessageBox.critical,
    MBOX_QUERY: qt.QMessageBox.question
}


signal_map = {
    SIGNAL_ACTIVATED: "triggered",
    SIGNAL_CLICKED: "clicked",
    SIGNAL_TOGGLED: "toggled",
    SIGNAL_SELECTED: "currentIndexChanged",
    SIGNAL_LISTVIEWSELECTED: "currentItemChanged",
    SIGNAL_LISTBOXSELECTED: "currentRowChanged",
    SIGNAL_CELLSELECTED: "currentCellChanged",
    SIGNAL_TABLECHANGED: "cellChanged",
    SIGNAL_TEXTCHANGED: "textChanged",
    SIGNAL_TEXTMODCHANGED: "modificationChanged",
    SIGNAL_EDITCHANGED: "textChanged",
    SIGNAL_VALUECHANGED: "valueChanged",
    SIGNAL_ENTER: "returnPressed",
    SIGNAL_TABSELECTED: "currentChanged",
    SIGNAL_NOTIFIER: "activated",
    SIGNAL_FINISHED: "finished",
    SIGNAL_BEFOREQUIT: "aboutToQuit",
}

event_map = {
    SIGNAL_BUTTONSELECTED: ("sig_buttonSelected", (int,)),
    SIGNAL_NUMEDITCHANGED: ("sig_numEditChanged", (int,)),
    SIGNAL_CELLCHANGED: ("sig_cellChanged", (int, int)),
    SIGNAL_TEXTSTATECHANGED: ("sig_textStateChanged", ()),  # using selectionChanged above does not work
    SIGNAL_QUERYCLOSE: ("sig_canClose", ()),
    SIGNAL_CLOSING: ("sig_closeEvent", ()),
    SIGNAL_SHOWN: ("sig_showEvent", ()),
    SIGNAL_HIDDEN: ("sig_hideEvent", ()),
    
    SIGNAL_FOCUS_IN: ("sig_focusInEvent", ()),
    SIGNAL_FOCUS_OUT: ("sig_focusOutEvent", ()),
    SIGNAL_LEFTCLICK: ("sig_leftClickEvent", ()),
    SIGNAL_RIGHTCLICK: ("sig_rightClickEvent", ()),
    SIGNAL_MIDDLECLICK: ("sig_middleClickEvent", ()),
    SIGNAL_LEFTDBLCLICK: ("sig_leftDblClickEvent", ()),
    SIGNAL_RIGHTDBLCLICK: ("sig_rightDblClickEvent", ()),
    SIGNAL_MIDDLEDBLCLICK: ("sig_middleDblClickEvent", ()),
}


# This is provided to support the custom API; it must be called before
# any widget classes are declared

# Note: args must be the types of arguments for the event handler (as
# in event_map above)

def add_custom_signal(signame, *args):
    # Importing all from defs above will have retrieved the custom signal by name
    sig = getattr(sys.modules[__name__], signame)
    event_name = signame.lower()
    event_map[sig] = (event_name, args)


def map_args(signal):
    return (
        (signal_map[signal], ()) if signal in signal_map else
        event_map[signal] if signal in event_map else
        None
    )


def map_attr(signal):
    return map_args(signal)[0]


# Ugly hacks to fix metaclass conflicts for widget classes
# and to allow auto-construction of event signals without
# a lot of boilerplate code in each widget class

QtMeta = type(qtc.QObject)


def setup_event_signals(bases, attrs, attrname='event_signals'):
    # This has to be done in a metaclass because new signals cannot be
    # dynamically added after a QObject subclass is created
    signal_iter = attrs.get(attrname, ())
    if attrname == 'event_signals':
        signal_iter = chain(
            getattr(PQtWidgetBase, attrname),
            signal_iter,
            *(getattr(base, attrname, ()) for base in bases)
        )
    for signal in signal_iter:
        signame, args = map_args(signal)
        attrs[signame] = qtc.pyqtSignal(*args, name=signame)


class PQtWidgetMeta(QtMeta):
    # Metaclass for non-sequence widgets
    
    def __new__(meta, name, bases, attrs):
        setup_event_signals(bases, attrs)
        setup_event_signals(bases, attrs, 'signals')  # for custom widgets
        return super(PQtWidgetMeta, meta).__new__(meta, name, bases, attrs)


# We need a different metaclass for sequence-like widgets
# because ABCMeta needs to be included

class PQtSequenceMeta(ABCMeta, QtMeta):
    # Metaclass for sequence widgets
    
    def __new__(meta, name, bases, attrs):
        setup_event_signals(bases, attrs)
        setup_event_signals(bases, attrs, 'signals')  # for custom widgets
        return super(PQtSequenceMeta, meta).__new__(meta, name, bases, attrs)
    
    def __init__(cls, name, bases, attrs):
        QtMeta.__init__(cls, name, bases, attrs)
        ABCMeta.__init__(cls, name, bases, attrs)


# NOTE: It would be nice to attach metaclasses as needed to the base
# classes below so we wouldn't have to declare them in each widget module
# in multiple class declarations. In PySide2 this works; however, in PyQt5
# when things are set up to resolve all metaclass conflicts, programs will
# not run without segfaulting with no useful error output. So there appears
# to be something in the bowels of PyQt5 that is preventing this from
# working, and which is not there in PySide2.


class PQtSignalBase(object):
    """Base class for common signal/slot functionality in Qt.
    """
    
    def connect_target(self, signal, target):
        sig = map_attr(signal)
        if sig is not None:
            try:
                getattr(self, sig).connect(target)
            except AttributeError:
                pass
    
    def do_notify(self, signal, *args):
        sig = map_attr(signal)
        if sig is not None:
            try:
                getattr(self, sig).emit(*args)
            except AttributeError:
                pass


class PQtActionMixin(object):
    
    def load_icon_from_data(self, data):
        px = qtg.QPixmap()
        px.loadFromData(data)
        return qtg.QIcon(px)
    
    def load_icon_from_file(self, filename):
        return qtg.QIcon(qtg.QPixmap(filename))
    
    def get_accel_str(self, key):
        return accel_map.get(key)


def qt_font_object(font_name, font_size=None, bold=None, italic=None):
    font = qtg.QFont(font_name)
    if font_size is not None:
        font.setPointSize(font_size)
    if bold is not None:
        font.setBold(bold)
    if italic is not None:
        font.setItalic(italic)
    return font


class PQtWidgetBase(PQtSignalBase):
    """Base class for common Qt widget methods.
    """
    
    fn_enable_get = 'isEnabled'
    fn_enable_set = 'setEnabled'
    
    fix_width_on_resize = False
    fix_height_on_resize = False
    
    widget_class = None
    
    event_signals = (SIGNAL_FOCUS_IN, SIGNAL_FOCUS_OUT)
    
    def update_widget(self):
        self.update()
    
    def preferred_width(self):
        return max(self.minimumSize().width(), self.sizeHint().width())
    
    def preferred_height(self):
        return max(self.minimumSize().height(), self.sizeHint().height())
    
    def get_width(self):
        return self.width()
    
    def get_height(self):
        return self.height()
    
    def set_size(self, width, height):
        self.resize(width, height)
    
    def get_left(self):
        return self.x()
    
    def get_top(self):
        return self.y()
    
    def set_position(self, left, top):
        self.move(left, top)
    
    def set_min_size(self, width, height):
        self.setMinimumSize(width, height)
    
    def set_min_width(self, width):
        self.setMinimumWidth(width)
    
    def set_min_height(self, height):
        self.setMinimumHeight(height)
    
    def fix_width(self, width):
        self.setMinimumWidth(width)
    
    def fix_height(self, height):
        self.setMinimumHeight(height)
    
    def set_width(self, width):
        super(PQtWidget, self).set_width(width)
        if self.fix_width_on_resize:
            self.fix_width(width)
    
    def set_height(self, height):
        super(PQtWidget, self).set_height(width)
        if self.fix_height_on_resize:
            self.fix_height(height)
    
    def _mapped_color(self, color):
        if isinstance(color, qtg.QColor):
            return color
        return color_map[color]
    
    def set_colors(self, fg=None, bg=None):
        palette = qtg.QPalette(self.palette())
        if fg is not None:
            palette.setColor(self.foregroundRole(), self._mapped_color(fg))
        if bg is not None:
            self.setAutoFillBackground(True)
            palette.setColor(self.backgroundRole(), self._mapped_color(bg))
        self.setPalette(palette)
    
    def set_foreground_color(self, color):
        self.set_colors(fg=color)
    
    def set_background_color(self, color):
        self.set_colors(bg=color)
    
    def get_font_name(self):
        return self.font().family()
    
    def get_font_size(self):
        return self.font().pointSize()
    
    def get_font_bold(self):
        return self.font().bold()
    
    def get_font_italic(self):
        return self.font().italic()
    
    def set_font_object(self, font_name, font_size, bold, italic):
        self.setFont(qt_font_object(font_name, font_size, bold, italic))
    
    def set_focus(self):
        self.setFocus()


class PQtMessageBox(PMessageBoxBase):
    """Customized Qt message box.
    """
    
    question_map = {
        ANSWER_YES: qt.QMessageBox.Yes,
        ANSWER_NO: qt.QMessageBox.No,
        ANSWER_CANCEL: qt.QMessageBox.Cancel,
        ANSWER_OK: qt.QMessageBox.Ok
    }
    
    def message_box(self, type, caption, text, default,
                    button1, button2=None, button3=None):
        
        buttons = button1
        if button2 is not None:
            buttons |= button2
            if button3 is not None:
                buttons |= button3
        
        return message_funcs[type](self.parent, caption, text,
                                   buttons, default)
    
    value_fn_map = {
        str: qt.QInputDialog.getText,
        int: qt.QInputDialog.getInt,
    }
    
    def enter_value(self, value_type, caption, prompt):
        f = self.value_fn_map.get(value_type)
        if not f:
            raise ValueError("Unsupported value type: {}".format(value_type))
        result, accepted = f(self.parent, caption, prompt)
        if accepted:
            return result


class PQtFileDialog(PFileDialogBase):
    
    def choose_directory(self, curdir):
        return str(qt.QFileDialog.getExistingDirectory(
            None, "Select Folder", curdir
        ))
    
    def _translate_filter_item(self, caption, ext_str):
        return "{} ({})".format(caption, ext_str)
    
    ext_sep = " "
    filter_sep = ";;"
    
    def _open_filename(self, path, filter, selected_filter):
        return qt.QFileDialog.getOpenFileName(self.parent, "Open",
                                              path, filter, selected_filter)[0]
    
    def _save_filename(self, path, filter, selected_filter):
        return qt.QFileDialog.getSaveFileName(self.parent, "Save",
                                              path, filter, selected_filter)[0]


class PQtAboutDialog(PAboutDialogBase):
    
    display_func = qt.QMessageBox.about


class PQtProgressDialog(PProgressDialogBase):
    
    def create_dialog(self, title, msg, max_num, stop_label, main_window):
        dlg = qt.QProgressDialog(msg, stop_label, 0, max_num, main_window)
        dlg.setWindowTitle(title)
        dlg.resize(len(title) * 14, dlg.height())  # kludge since qt doesn't automatically resize to the title
        dlg.setMinimumDuration(0)
        dlg.setWindowModality(qtc.Qt.WindowModal)
        self.parent.app.processEvents()
        return dlg
    
    def show_dialog(self):
        self.dialog.show()
        self.parent.app.processEvents()
    
    def was_canceled(self):
        return self.dialog.wasCanceled()
    
    def update_progress(self, num):
        self.dialog.setValue(num)
        self.parent.app.processEvents()
    
    def complete(self):
        self.dialog.close()
        self.parent.app.processEvents()
    
    def close_dialog(self):
        self.dialog.close()


# NOTE: For some reason, PyQt5 insists on having the Qt classes first in the MRO,
# so all MROs have been switched around to do that. PySide2 does not require this.
# Note sure why the difference between the two, possibly different wrapper generation
# tools.


class PTopWindow(qt.QMainWindow, PQtWidgetBase, PTopWindowBase,
                 metaclass=PQtWidgetMeta):
    """Customized Qt top window class.
    """
    
    widget_class = qt.QMainWindow
    
    event_signals = (SIGNAL_QUERYCLOSE, SIGNAL_CLOSING, SIGNAL_SHOWN, SIGNAL_HIDDEN)
    
    def __init__(self, manager):
        qt.QMainWindow.__init__(self)
        PTopWindowBase.__init__(self, manager)
        self._show_state = False  # to avoid multiple firings of shown/hidden events
    
    def set_client_widget(self, client_widget):
        PTopWindowBase.set_client_widget(self, client_widget)
        self.setCentralWidget(client_widget)
    
    def set_caption(self, caption):
        self.setWindowTitle(caption)
    
    def get_desktop_rect(self, primary=True):
        # Correctly handle virtual desktop across multiple screens
        desktop = self.app.desktop()
        l = desktop.x()
        t = desktop.y()
        w = desktop.width()
        h = desktop.height()
        if desktop.isVirtualDesktop() and primary:
            # Default to centering on the primary screen
            i = desktop.primaryScreen()
            n = desktop.numScreens()
            w = w / n
            # NOTE: We have to check for i > 0 here because in some
            # cases (e.g., when running in a VirtualBox), Qt thinks
            # the desktop is "virtual" but there's only one screen and
            # desktop.primaryScreen returns 0 instead of 1.
            if i > 0:
                l += w * (i - 1)
        return l, t, w, h
    
    def size_maximize(self):
        if self.shown:
            self.showMaximized()
        else:
            self._showMax = True
    
    def get_frame_size(self):
        return self.frameGeometry().width(), self.frameGeometry().height()
    
    def show_init(self):
        PTopWindowBase.show_init(self)
        if hasattr(self, '_showMax'):
            self.showMaximized()
            del self._showMax
        else:
            qt.QMainWindow.show(self)
    
    def do_exit(self):
        self.close()
    
    def closeEvent(self, event):
        self.do_notify(SIGNAL_QUERYCLOSE)  # a handler here can adjust state to be checked below
        
        # Note that Qt signals/slots are by default synchronous (and we don't use queued connections),
        # so ``do_notify`` will cause all handlers to execute before it returns
        if self.can_close():
            self.do_notify(SIGNAL_CLOSING)
            event.accept()
        else:
            event.ignore()
    
    def showEvent(self, event):
        if not self._show_state:
            self.do_notify(SIGNAL_SHOWN)
            self._show_state = True
    
    def hideEvent(self, event):
        if self._show_state:
            self.do_notify(SIGNAL_HIDDEN)
            self._show_state = False
    
    def set_iconfile(self, iconfile):
        self.setWindowIcon(qtg.QIcon(qtg.QPixmap(iconfile)))
    
    def get_current_geometry(self):
        p = self.pos()
        s = self.size()
        return p.x(), p.y(), s.width(), s.height()


QtAppMeta = type(qt.QApplication)


class PQtAppMeta(QtAppMeta):
    # Metaclass for application object
    
    def __new__(meta, name, bases, attrs):
        setup_event_signals(bases, attrs, 'signals')
        return super(PQtAppMeta, meta).__new__(meta, name, bases, attrs)


class PApplication(qt.QApplication, PQtSignalBase, PApplicationBase,
                   metaclass=PQtAppMeta):
    """Customized Qt application class.
    """
    
    about_dialog_class = PQtAboutDialog
    message_box_class = PQtMessageBox
    file_dialog_class = PQtFileDialog
    progress_dialog_class = PQtProgressDialog
    
    def __init__(self, arglist=[]):
        PApplicationBase.__init__(self, arglist)
        qt.QApplication.__init__(self, arglist)
        
        self.do_create()  # Qt allows widget creation in constructor
        
        self.about_toolkit_func = self.aboutQt
    
    def setup_signals(self):
        # 'automagic' signal connection
        self.setup_notify(SIGNAL_BEFOREQUIT, self.before_quit)
        
        PApplicationBase.setup_signals(self)
    
    def event_loop(self):
        self.exec_()
    
    def process_events(self):
        self.processEvents()



#!/usr/bin/env python3
"""
Module IMAGEVIEW -- UI Image View Widget
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""


from .app import PWidgetBase


class PImageViewBase(PWidgetBase):
    
    def __init__(self, manager, parent, filename=None, geometry=None):
        PWidgetBase.__init__(self, manager, parent, geometry=geometry)
        self.filename = filename
        if filename:
            self.load_from_file(filename)
    
    supported_formats = None  # must implement in toolkit subclass as static method
    
    def load_from_file(self, filename):
        # Must implement in toolkit subclass
        raise NotImplementedError
    
    def get_image_size(self):
        # Must implement in toolkit subclass
        raise NotImplementedError
    
    def zoom_to(self, width, height):
        # Must implement in toolkit subclass
        raise NotImplementedError
    
    def zoom_by(self, factor):
        w, h = self.get_image_size()
        self.zoom_to(w * factor, h * factor)
    
    def fit_to_window(self):
        w, h = self.get_size()
        self.zoom_to(w, h)
    
    def rotate_90(self, clockwise):
        # Must implement in toolkit subclass
        raise NotImplementedError
    
    def rotate_left(self):
        self.rotate_90(False)
    
    def rotate_right(self):
        self.rotate_90(True)



#!/usr/bin/env python3
"""
Module DIALOGS -- UI Dialog Runner and Standard Dialogs
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.app import PDialog
from plib.ui.base.app import PControllerBase
from plib.ui.common import set_bold
from plib.ui.widgets import *


class DialogRunner(PControllerBase):
    
    dialog_class = PDialog
    
    dialog_client = None
    
    def __init__(self, manager, caption, callback=None):
        PControllerBase.__init__(self)
        self.manager = manager
        self.caption = caption
        self.callback = callback
        
        self.dialog = None
    
    def create_widgets(self):
        self.dialog = self.dialog_class(self, self.app.main_window, self.caption, self.dialog_client)
    
    def run(self):
        self.do_create()
        self.dialog.setup_notify(SIGNAL_FINISHED, self.dialog_done)
        self.dialog.display()
    
    show_dialog = run  # syntactic sugar
    
    def dialog_done(self, accepted):
        if accepted and self.callback:
            self.callback(self.get_result())
        self.dialog = None
    
    def get_result(self):
        return None


class DisplayDialog(DialogRunner):
    
    def __init__(self, manager, caption, client=None):
        DialogRunner.__init__(self, manager, caption)
        if client:
            self.dialog_client = client


class StringSelectDialog(DialogRunner):
    
    def __init__(self, manager, caption, values, prompt=None, starting_value=None, callback=None, font=None):
        DialogRunner.__init__(self, manager, caption, callback)
        self.values = values
        self.prompt = prompt
        self.starting_value = starting_value
        self.font = font
        self.label_font = set_bold(font)
    
    @property
    def dialog_client(self):
        return frame(ALIGN_JUST, LAYOUT_VERTICAL, [
            panel(ALIGN_JUST, LAYOUT_VERTICAL, ([
                box(ALIGN_TOP, LAYOUT_VERTICAL, [
                    label('current', self.prompt, font=self.label_font),
                ])] if self.prompt else []) + [
                listbox('values', [], font=self.font),
            ]),
            panel(ALIGN_BOTTOM, LAYOUT_HORIZONTAL, [
                action_button(ACTION_OK, font=self.font),
                action_button(ACTION_CANCEL, font=self.font),
            ]),
        ])
    
    def populate_data(self):
        self.listbox_values.extend(self.values)
        if self.starting_value:
            self.listbox_values.set_current_text(self.starting_value)
    
    def get_result(self):
        return self.listbox_values.current_text()



#!/usr/bin/env python3
"""
Module WX.GROUP -- Python wxWidgets Button Group Objects
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python GUI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the wxWidgets UI objects for the panel widgets.
"""

import wx

from plib.ui.defs import *
from plib.ui.base.group import PButtonGroupMixin

from .app import PWxSequenceWidget, eventManager, wx_font_object
from .form import PPanel


class PButtonGroupBase(PWxSequenceWidget, PButtonGroupMixin, PPanel):
    
    button_class = None
    
    def __init__(self, manager, parent, align, layout, items=None, value=None, starting_index=None, font=None,
                 style=PANEL_PLAIN, margin=None, spacing=None):
        
        self.clearing = False
        PPanel.__init__(self, manager, parent, align, layout,
                        style=style, margin=margin, spacing=spacing)
        PButtonGroupMixin.__init__(self, items, value, starting_index, font)
    
    def complete_init(self, items, value, starting_index):
        PButtonGroupMixin.complete_init(self, items, value, starting_index)
        self.SetSizerAndFit(self._sizer)
    
    def init_group(self):
        self._buttons = []
    
    def add_items(self, items, value=None, starting_index=None):
        PButtonGroupMixin.add_items(self, items, value, starting_index)
        self.SetSizerAndFit(self._sizer)
    
    button_style = 0
    
    def create_button(self, item):
        return self.button_class(self, label=item, style=self.button_style)
    
    def set_button_font(self, button, font_name=None, font_size=None, bold=None, italic=None):
        font = wx_font_object(font_name, font_size, bold, italic)
        button.SetFont(font)
    
    def init_button(self, button):
        self._buttons.append(button)
        eventManager.Register(self.buttonEvent, self.button_event, button)
    
    def group_buttons(self):
        return self._buttons  # TODO: can this just be self.contents?
    
    def del_button(self, button):
        self._buttons.remove(button)
    
    def remove_widget(self, widget):
        self._sizer.Detach(widget)
        if not self.clearing:
            self.SetSizerAndFit(self._sizer)
    
    def clear(self):
        if self._buttons:
            self.clearing = True
            PButtonGroupMixin.clear(self)
            self.SetSizerAndFit(self._sizer)
            self.clearing = False
    
    def is_checked(self, button):
        return button.GetValue()
    
    def check_button(self, button, checked=True):
        button.SetValue(checked)
        # This event doesn't get fired by wx when the button is checked in code,
        # so we have to emulate it manually
        self.on_button_toggled(button, checked)
    
    def button_text(self, button):
        return button.GetLabel()
    
    def set_button_text(self, button, value):
        button.SetLabel(value)


class PButtonGroup(PButtonGroupBase):
    
    button_class = wx.ToggleButton
    button_event = wx.EVT_TOGGLEBUTTON
    
    def buttonEvent(self, event):
        # Here we have to go around Robin Hood's barn because
        # wx has no built-in support for grouping toggle buttons (??)
        button = event.GetEventObject()
        checked = event.IsChecked()  # or button.GetValue()
        if checked:
            for other_button in self.buttons:
                if other_button is not button:
                    other_button.SetValue(False)
        self.on_button_toggled(button, checked)


class PRadioGroup(PButtonGroupBase):
    
    button_class = wx.RadioButton
    button_event = wx.EVT_RADIOBUTTON
    
    next_button_style = wx.RB_GROUP
    
    @property
    def button_style(self):
        style = self.next_button_style
        self.next_button_style = 0
        return style
    
    def buttonEvent(self, event):
        # Here this works simply because wx radio buttons can be grouped in the style
        button = event.GetEventObject()
        checked = event.IsChecked()  # or button.GetValue()
        self.on_button_toggled(button, checked)



#!/usr/bin/env python3
"""
Module TEXTFILE -- UI File-Aware Text Control Helper
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import os

from plib.ui.defs import *
from plib.ui.fileaware import PFileAware


class PTextFile(PFileAware):
    """Make a text edit control file-aware.
    """
    
    def __init__(self, app, control, filter=None, filter_index=0, file_path=None, update_filter=True, update_index=0, update_path=True):
        PFileAware.__init__(self, app, filter, filter_index, file_path, update_filter, update_index, update_path)
        self.control = control  # should normally be a PEditControl
        self.dirty = False
        
        # So we can keep track of "dirty" state
        control.setup_notify(SIGNAL_TEXTMODCHANGED, self.editor_changed)
    
    @property
    def filesize(self):
        return len(self.control.edit_text)
    
    untitled = object()
    
    def new_file(self):
        super(PTextFile, self).new_file()
        self.set_filename(self.untitled)
    
    def open_data(self, filename):
        with open(filename, 'r') as f:
            data = f.read()
        self.control.edit_text = data
        self.dirty = False
    
    def editor_changed(self, changed):
        self.dirty = changed
    
    def save_data(self, filename):
        data = self.control.edit_text
        with open(filename, 'w') as f:
            f.write(data)
        self.dirty = False
    
    def filename_unknown(self):
        return (not self.filename) or (self.filename is self.untitled)
    
    def close_file(self):
        # Calling application must make sure data is saved first
        self.control.clear_edit()
        self.dirty = False
        super(PTextFile, self).close_file()



#!/usr/bin/env python3
"""
Module WX.SOCKET -- Python Wx Socket Notifier Objects
Sub-Package UI.TOOLKITS.WX of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI socket notifier objects.
"""

import select

from plib.stdlib.classtools import Singleton

from plib.ui.defs import *
from plib.ui.base.socket import PSocketNotifierBase

from .app import PApplication


# NOTE: We don't use the higher-level selectors module here
# because we want separate registration of readers and writers,
# even if both are for the same underlying socket object

class PWxSocketManager(Singleton):
    
    timeout = 0.1
    
    def _init(self):
        self.readers = []
        self.writers = []
        self.exit_loop = False
    
    @property
    def app(self):
        return PApplication.Get()
    
    def add_reader(self, reader):
        self.readers.append(reader)
    
    def add_writer(self, writer):
        self.writers.append(writer)
    
    def remove_reader(self, reader):
        self.readers.remove(reader)
    
    def remove_writer(self, writer):
        self.writers.remove(writer)
    
    def run_select_loop(self,
                        select=select.select):
        
        self.exit_loop = False
        timeout = self.timeout
        e = []  # no socket exception handlers
        while (not self.exit_loop) and (self.readers or self.writers):
            r, w, e = select(self.readers, self.writers, e, timeout)
            for reader in r:
                reader.process_read()
            for writer in w:
                writer.process_write()
            # This is a kludge, but wx gives us no way to have its own event loop
            # process socket events, so it's the best we can do
            self.app.process_events()
    
    @classmethod
    def get_instance(cls):
        return cls.__dict__.get("__inst__")


class PSocketNotifier(PSocketNotifierBase):
    
    manager = PWxSocketManager()
    
    notify_map = {
        NOTIFY_READ: 'add_reader',
        NOTIFY_WRITE: 'add_writer',
    }
    
    disable_map = {
        NOTIFY_READ: 'remove_reader',
        NOTIFY_WRITE: 'remove_writer',
    }
    
    def __init__(self, obj, notify_type, select_fn, notify_fn):
        PSocketNotifierBase.__init__(self, obj, notify_type, select_fn, notify_fn)
        
        process_fn = self.process_fn
        null_fn = (lambda: None)
        self.process_read = process_fn if (notify_type == NOTIFY_READ) else null_fn
        self.process_write = process_fn if (notify_type == NOTIFY_WRITE) else null_fn
        
        self.set_enabled(True)
    
    def fileno(self):
        # We need this so select will recognize us as a file-like object
        return self._obj.fileno()
    
    def process_fn(self):
        if self.select_fn():
            self.notify_fn(self._obj)
    
    def set_enabled(self, enable):
        method_map = (self.disable_map, self.notify_map)[enable]
        method = getattr(self.manager, method_map[self._notify_type])
        method(self)
    
    @classmethod
    def start(cls):
        inst = cls.manager.get_instance()
        inst.run_select_loop()
    
    @classmethod
    def done(cls):
        inst = cls.manager.get_instance()
        inst.exit_loop = True



#!/usr/bin/env python3
"""
Module MOUSE -- UI Mouse Widget Mixin
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from PyQt5.QtCore import Qt

from plib.ui.defs import *
from plib.ui.base.mouse import PMouseBase


class PMouseMixin(PMouseBase):
    
    event_signals = (
        SIGNAL_LEFTCLICK,
        SIGNAL_RIGHTCLICK,
        SIGNAL_MIDDLECLICK,
        SIGNAL_LEFTDBLCLICK,
        SIGNAL_RIGHTDBLCLICK,
        SIGNAL_MIDDLEDBLCLICK,
    )
    
    button_events = {
        Qt.LeftButton: SIGNAL_LEFTCLICK,
        Qt.RightButton: SIGNAL_RIGHTCLICK,
        Qt.MiddleButton: SIGNAL_MIDDLECLICK,
    }
    
    double_events = {
        Qt.LeftButton: SIGNAL_LEFTDBLCLICK,
        Qt.RightButton: SIGNAL_RIGHTDBLCLICK,
        Qt.MiddleButton: SIGNAL_MIDDLEDBLCLICK,
    }
    
    def do_button_notify(self, event_map, event):
        sig = event_map.get(event.button())
        if sig:
            self.do_notify(sig)
    
    button_pressed = None
    
    def mousePressEvent(self, event):
        self.button_pressed = event.button()
    
    def mouseReleaseEvent(self, event):
        if (self.button_pressed is not None) and (event.button() == self.button_pressed):
            self.do_button_notify(self.button_events, event)
        self.button_pressed = None
    
    def mouseDoubleClickEvent(self, event):
        self.do_button_notify(self.double_events, event)



#!/usr/bin/env python3
"""
Module APP -- UI Application Classes
Sub-Package UI of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import sys

from plib.ui.imp import get_toolkit_object


# We will always need this one so we automatically import it

PApplication = get_toolkit_object('app', 'PApplication')


# This also serves as a nice reference for what is where

TOOLKIT_MODS = dict(
    # Standard widgets
    PButton='button',
    PActionButton='button',
    PCheckBox='button',
    PComboBox='combo',
    PNumComboBox='combo',
    PSortedComboBox='combo',
    PTextDisplay='display',
    PEditBox='editctrl',
    PNumEditBox='editctrl',
    PSpinEditBox='editctrl',
    PEditControl='editctrl',
    PPanel='form',
    PGroupBox='groupbox',
    PButtonGroup='group',
    PRadioGroup='group',
    PHtmlDisplay='html',
    PImageView='imageview',
    PTextLabel='label',
    PListBox='listbox',
    PSortedListBox='listbox',
    PTreeView='listview',
    PListView='listview',
    PSortedListView='listview',
    PPageWidget='pagewidget',
    PScroller='scroller',
    PScrollingWrapper='scroller',
    PTable='table',
    PTabWidget='tabwidget',
    
    # Special widgets
    PContainer='container',
    
    # Dialog
    PDialog='dialog',
    
    # Special helpers
    PSocketNotifier='socket',
    PTableLabels='table',
    
    # Mixins for custom widgets to add signals
    PFocusMixin='focus',
    #PKeyboardMixin='keyboard',
    PMouseMixin='mouse',
)


if sys.version_info[:2] >= (3, 7):
    # PEP 562 functionality is available
    
    mod_names = dir(sys.modules[__name__])
    
    
    def __getattr__(name):
        # We do this so that the only toolkit objects that actually get imported
        # are the ones that the application actually uses
        modname = TOOLKIT_MODS.get(name)
        if not modname:
            raise AttributeError("Toolkit object {} not found".format(name))
        obj = get_toolkit_object(modname, name)
        # This ensures we don't have to call this function again for this name
        setattr(sys.modules[__name__], name, obj)
        return obj
    
    
    def __dir__():
        return sorted(mod_names + list(TOOLKIT_MODS))


else:
    # PEP 562 functionality is not available so we have to roll our own
    
    from plib.stdlib.util import ModuleProxy
    
    def make_proxy(modname, name):
        return lambda: get_toolkit_object(modname, name)
    
    proxy_names = {
        name: make_proxy(modname, name)
        for name, modname in TOOLKIT_MODS.items()
    }
    
    ModuleProxy(__name__).init_proxy(__name__, None, globals(), locals(),
                                     names=proxy_names, autodiscover=False)



#!/usr/bin/env python3
"""
Module QT5.SOCKET -- Python Qt 5 Socket Notifier Objects
Sub-Package UI.TOOLKITS.QT5 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI socket notifier objects.
"""

from PyQt5 import QtCore as qtc

from plib.ui.defs import *
from plib.ui.base.app import PSignalBase
from plib.ui.base.socket import PSocketNotifierBase

from .app import PQtSignalBase


notify_types = {
    NOTIFY_READ: qtc.QSocketNotifier.Read,
    NOTIFY_WRITE: qtc.QSocketNotifier.Write
}


class PSocketNotifier(qtc.QSocketNotifier, PQtSignalBase, PSignalBase, PSocketNotifierBase):
    
    auto_enable = True
    
    def __init__(self, obj, notify_type, select_fn, notify_fn):
        PSocketNotifierBase.__init__(self, obj, notify_type, select_fn, notify_fn)
        qtc.QSocketNotifier.__init__(self, obj.fileno(),
                                     notify_types[notify_type])
        self.setup_notify(SIGNAL_NOTIFIER, self.handle_notify)
    
    def set_enabled(self, enable):
        self.setEnabled(enable)
    
    # We don't need to override the start or done class methods since Qt's GUI
    # event loop already handles socket events, so just instantiating and enabling
    # the socket notifier is enough
    
    def handle_notify(self, sock):
        self.set_enabled(False)
        if (sock == self._obj.fileno()) and self.select_fn():
            self.notify_fn(self._obj)
        if self.auto_enable:
            self.set_enabled(True)



#!/usr/bin/env python3
"""
PXMLVIEW.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A simple read-only XML file viewer.
"""

import sys
import os
import io
import xml.etree.ElementTree as etree

from plib.ui import __version__
from plib.ui.defs import *
from plib.ui.app import PApplication, PTreeView
from plib.ui.fileaware import PFileAware
from plib.ui.widgets import *


def get_xml_node_tree(node):
    # Translate node into form suitable for tree view widget
    cols = (
        node.tag,
        ", ".join("{}={}".format(key, value) for key, value in node.items()),
        node.text,
        node.tail,
    )
    children = [get_xml_node_tree(child) for child in list(node)]
    return (cols, children)


def get_xml_tree(filename):
    # Parse XML file into a tree of PTreeView data
    xml = etree.parse(filename)
    return [get_xml_node_tree(xml.getroot())]


class XMLFileLoader(PFileAware):
    
    def __init__(self, app):
        PFileAware.__init__(self, app, [
            ("XML Files", ["xml"]),
            ("All Files", ["*"]),
        ], file_path=app.main_path)
    
    def open_data(self, filename):
        self.app.add_file(filename, get_xml_tree(filename))
        self.close_file()  # to be ready for the next load
    
    def save_data(self, filename):
        pass  # this app does not edit/save files


class XMLViewer(PApplication):
    
    about_data = {
        'name': "XMLViewer",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "XML File Viewer",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net",
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    menu_actions = [
        (MENU_FILE, (ACTION_FILE_OPEN, ACTION_EXIT,)),
        (MENU_HELP, (ACTION_ABOUT, ACTION_ABOUT_TOOLKIT,)),
    ]
    
    toolbar_actions = [
        (ACTION_FILE_OPEN,),
        (ACTION_ABOUT, ACTION_ABOUT_TOOLKIT, ACTION_EXIT,),
    ]
    
    main_path = os.path.split(os.path.realpath(__file__))[0]
    
    main_title = "XML File Viewer"
    main_iconfile = os.path.join(main_path, "pxmlview.png")
    
    main_size = SIZE_MAXIMIZED
    large_icons = True
    
    main_widget = tabwidget('files', [])
    
    view_class = PTreeView
    view_header = ["Tag", "Attrs", "Text", "Tail"]
    view_font_size = 16 if sys.platform == 'darwin' else 12
    
    filenames_to_open = None
    
    def check_file_open(self, filename):
        if os.path.isfile(filename):
            print("Opening", filename)
            self.loader.open_filename(filename)
        else:
            print(filename, "not found!")
            self.filenames_not_found.append(filename)
    
    def after_create(self):
        self.filenames = []
        self.loader = XMLFileLoader(self)
        self.filenames_not_found = []
        for filename in (self.filenames_to_open or ()):
            self.check_file_open(filename)
    
    def add_file(self, filename, data):
        self.filenames.append(filename)
        font_size = self.view_font_size
        view = self.view_class(self, self.tabwidget_files, labels=self.view_header, data=data, auto_expand=True,
                               font=("Arial", font_size), header_font=("Arial", font_size, True))
        self.tabwidget_files.append_and_focus((os.path.basename(filename), view))
    
    def on_app_shown(self):
        if self.filenames_not_found:
            self.message_box.error("File Error", "Files {} not found".format(self.filenames_not_found))
    
    def on_file_open(self):
        self.loader.open_file()
    
    def set_status_text(self, text):
        print(text)
        self.main_window.statusbar.set_text(text)
    
    def on_files_selected(self, index):
        self.set_status_text("Viewing file: {}".format(self.filenames[index]))
    
    def on_about(self):
        self.about()
    
    def on_about_toolkit(self):
        self.about_toolkit()
    
    def on_exit(self):
        self.exit_app()


if __name__ == "__main__":
    if len(sys.argv) > 1:
        XMLViewer.filenames_to_open = sys.argv[1:]
    XMLViewer().run()



#!/usr/bin/env python3
"""
Module PYSIDE2.SCROLLER -- Python PySide 2 Scroller Objects
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the PySide 2 UI objects for the panel widgets.
"""

from PySide2 import QtWidgets as qt

from plib.ui.base.scroller import PScrollerBase, PScrollingWrapperBase

from .app import PQtWidget


class PQtScrollerBase(PQtWidget, qt.QScrollArea):
    
    def __init__(self, parent):
        qt.QScrollArea.__init__(self, parent)
    
    def set_child(self, child):
        self.setWidget(child)


class PScroller(PQtScrollerBase, PScrollerBase):
    
    def __init__(self, manager, parent, child=None):
        PQtScrollerBase.__init__(self, parent)
        PScrollerBase.__init__(self, manager, parent, child)


class PScrollingWrapper(PQtScrollerBase, PScrollingWrapperBase):
    
    def __init__(self, manager, parent, child_spec):
        PQtScrollerBase.__init__(self, parent)
        self.setWidgetResizable(True)
        PScrollingWrapperBase.__init__(self, manager, parent, child_spec)
    
    def get_layout(self, child):
        # Intended for child to be a PPanel or subclass
        layout = getattr(child, '_playout', None)
        if layout is None:
            return None
        return layout
    
    def setup_layout(self, layout):
        layout.setSizeConstraint(layout.SetMinAndMaxSize)



#!/usr/bin/env python3
"""
Module MOUSE -- UI Mouse Widget Mixin
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from plib.ui.defs import *


class PMouseBase:
    
    signals = (
        SIGNAL_LEFTCLICK,
        SIGNAL_RIGHTCLICK,
        SIGNAL_MIDDLECLICK,
        SIGNAL_LEFTDBLCLICK,
        SIGNAL_RIGHTDBLCLICK,
        SIGNAL_MIDDLEDBLCLICK,
    )



#!/usr/bin/env python3
"""
Module HTML -- UI Html Display Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

from .app import PWidgetBase


class PHtmlDisplayBase(PWidgetBase):
    """Base class for widget that displays html.
    """
    
    def __init__(self, manager, parent, html="",
                 scrolling=False, font=None):
        
        PWidgetBase.__init__(self, manager, parent,
                             font=font)
        self.setup_scrolling(scrolling)
        self.set_html(html)
    
    def setup_scrolling(self, scrolling):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def get_html(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def set_html(self, value):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def clear_html(self):
        # This can be overridden in toolkits
        self.set_html("")



#!/usr/bin/env python3
"""
Module PYSIDE2.LISTBOX-- Python PySide 2 List Box Objects
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the PySide 2 UI objects for the tree/list view widgets.
"""

from PySide2 import QtWidgets as qt

from plib.ui.base.listbox import PListBoxBase, PSortedListBoxBase

from .app import PQtSequenceWidget


class PQtListBoxBase(PQtSequenceWidget, qt.QListWidget):
    
    fix_width_on_resize = True
    
    def __init__(self, parent):
        qt.QListWidget.__init__(self, parent)
        self.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
    
    #def item_height(self, index):
    #    return self.itemWidget(self.item(index)).height()
    
    def current_index(self):
        return self.currentRow()
    
    def set_current_index(self, index):
        self.setCurrentRow(index)
    
    def count(self, value):
        # Method name collision, we want it to be the Python sequence
        # count method here.
        return PListBoxBase.count(self, value)
    
    def _indexlen(self):
        # Let this method access the Qt combo box count method.
        return qt.QListWidget.count(self)
    
    def _get_data(self, index):
        return str(self.item(index).text())
    
    def _set_data(self, index, value):
        self.item(index).setText(str(value))
    
    def _add_data(self, index, value):
        self.insertItem(index, str(value))
    
    def _del_data(self, index):
        self.takeItem(index)


class PListBox(PQtListBoxBase, PListBoxBase):
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None, font=None):
        
        PQtListBoxBase.__init__(self, parent)
        PListBoxBase.__init__(self, manager, parent, items, value=value,
                              geometry=geometry, font=font)


class PSortedListBox(PQtListBoxBase, PSortedListBoxBase):
    
    def __init__(self, manager, parent, items=None, value=None,
                 geometry=None, font=None, key=None):
        
        PQtListBoxBase.__init__(self, parent)
        PSortedListBoxBase.__init__(self, manager, parent, items, value=value,
                                    geometry=geometry, font=font, key=key)



#!/usr/bin/env python3
"""
Module QT5.IMAGEVIEW -- Python Qt 5 Image View Objects
Sub-Package UI.TOOLKITS.PYSIDE2 of Package PLIB3 -- Python UI Toolkits
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

This module contains the Qt 5 UI objects for the panel widgets.
"""

from PyQt5 import QtCore as qtc, QtGui as qtg, QtWidgets as qt

from plib.ui.base.imageview import PImageViewBase

from .app import PQtWidgetMeta, PQtWidgetBase


class PImageView(qt.QScrollArea, PQtWidgetBase, PImageViewBase,
                 metaclass=PQtWidgetMeta):
    
    image_view = None
    
    def __init__(self, manager, parent, filename=None, geometry=None):
        qt.QScrollArea.__init__(self, parent)
        self.setAlignment(qtc.Qt.AlignTop | qtc.Qt.AlignLeft)
        PImageViewBase.__init__(self, manager, parent, filename=filename, geometry=geometry)
    
    @staticmethod
    def supported_formats():
        # This must not be called until after the application object is created
        return [bytes(f).decode() for f in qtg.QImageReader.supportedImageFormats()]
    
    def load_from_file(self, filename):
        pixmap = qtg.QPixmap(filename)
        self.image_view = qt.QLabel()
        self.image_view.setAlignment(qtc.Qt.AlignTop | qtc.Qt.AlignLeft)
        self.image_view.setPixmap(pixmap)
        self.setWidget(self.image_view)
    
    def get_image_size(self):
        size = self.image_view.pixmap().size()
        return size.width(), size.height()
    
    def zoom_to(self, width, height):
        pixmap = self.image_view.pixmap()
        self.image_view.setPixmap(pixmap.scaled(width, height, qtc.Qt.KeepAspectRatio))
        self.image_view.resize(width, height)
    
    def rotate_90(self, clockwise):
        width, height = self.get_image_size()
        transform = qtg.QTransform()
        transform.rotate(90.0 * (1.0 if clockwise else -1.0))
        pixmap = self.image_view.pixmap()
        self.image_view.setPixmap(pixmap.transformed(transform))
        self.image_view.resize(height, width)



#!/usr/bin/env python3
"""
PIMGVIEW.PY
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information

A simple read-only image file viewer.
"""

import sys
import os

from plib.ui import __version__
from plib.ui.defs import *
from plib.ui import common
from plib.ui.app import PApplication, PImageView
from plib.ui.fileaware import PFileAware
from plib.ui.widgets import *


IMAGE_EXTENSIONS = []


class ImageFileLoader(PFileAware):
    
    def __init__(self, app):
        PFileAware.__init__(self, app, [
            ("Image Files", list(IMAGE_EXTENSIONS)),
        ] + [
            ("All Files", ["*"]),
        ], file_path=app.main_path)
    
    def open_data(self, filename):
        self.app.add_file(filename)
        self.close_file()  # to be ready for the next load
    
    def save_data(self, filename):
        pass  # this app does not edit/save files


# Add custom actions
ACTION_FIT = ACTION_REFRESH + 20
ACTION_ZOOM_IN = ACTION_REFRESH + 21
ACTION_ZOOM_OUT = ACTION_REFRESH + 22
ACTION_ROTATE_LEFT = ACTION_REFRESH + 23
ACTION_ROTATE_RIGHT = ACTION_REFRESH + 24

common.add_action(
    ACTION_FIT,
    'fit', "&Fit", "Fit image to window", "apply"
)

common.add_action(
    ACTION_ZOOM_IN,
    'zoom_in', "Zoom &In", "Zoom image in", "add"
)

common.add_action(
    ACTION_ZOOM_OUT,
    'zoom_out', "Zoom &Out", "Zoom image out", "remove"
)


def custom_icon(name):
    return os.path.join(os.path.dirname(__file__), "{}.png".format(name))


common.add_action(
    ACTION_ROTATE_LEFT,
    'rotate_left', "Rotate &Left", "Rotate image 90 degrees counterclockwise", custom_icon("rotate_left")
)

common.add_action(
    ACTION_ROTATE_RIGHT,
    'rotate_right', "Rotate &Right", "Rotate image 90 degrees clockwise", custom_icon("rotate_right")
)


class ImageViewer(PApplication):
    
    about_data = {
        'name': "ImageViewer",
        'version': "{} on Python {}".format(
            __version__,
            sys.version.split()[0]
        ),
        'description': "Image Viewer",
        'copyright': "Copyright (C) 2008-2022 by Peter A. Donis",
        'license': "GNU General Public License (GPL) Version 2",
        'developer': "Peter Donis",
        'website': "http://www.peterdonis.net",
    }
    
    about_format = "{name} {version}\n\n{description}\n\n{copyright}\n{license}\n\nDeveloped by {developer}\n{website}"
    
    menu_actions = [
        (MENU_FILE, (ACTION_FILE_OPEN, ACTION_EXIT,)),
        (MENU_ACTION, (ACTION_FIT, ACTION_ZOOM_IN, ACTION_ZOOM_OUT, ACTION_ROTATE_LEFT, ACTION_ROTATE_RIGHT)),
        (MENU_HELP, (ACTION_ABOUT, ACTION_ABOUT_TOOLKIT,)),
    ]
    
    toolbar_actions = [
        (ACTION_FILE_OPEN,),
        (ACTION_FIT, ACTION_ZOOM_IN, ACTION_ZOOM_OUT, ACTION_ROTATE_LEFT, ACTION_ROTATE_RIGHT),
        (ACTION_ABOUT, ACTION_ABOUT_TOOLKIT, ACTION_EXIT,),
    ]
    
    main_path = os.path.split(os.path.realpath(__file__))[0]
    
    main_title = "Image Viewer"
    main_iconfile = os.path.join(main_path, "pimgview.png")
    
    main_size = SIZE_MAXIMIZED
    
    main_widget = tabwidget('images', [])
    
    view_class = PImageView
    
    filenames_to_open = None
    
    def check_file_open(self, filename):
        if os.path.isfile(filename):
            print("Opening", filename)
            self.loader.open_filename(filename)
        else:
            print(filename, "not found!")
            self.filenames_not_found.append(filename)
    
    def after_create(self):
        # Call this here for safety since some toolkits won't support it until after the app object is created
        IMAGE_EXTENSIONS.extend(sorted(self.view_class.supported_formats()))
        self.filenames = []
        self.views = []
        self.loader = ImageFileLoader(self)
        self.filenames_not_found = []
        for filename in (self.filenames_to_open or ()):
            self.check_file_open(filename)
    
    def add_file(self, filename):
        self.filenames.append(filename)
        view = self.view_class(self, self.tabwidget_images, filename)
        self.views.append(view)
        self.tabwidget_images.append_and_focus((os.path.basename(filename), view))
    
    def on_app_shown(self):
        if self.filenames_not_found:
            self.message_box.error("File Error", "Files {} not found".format(self.filenames_not_found))
    
    def on_file_open(self):
        self.loader.open_file()
    
    def set_status_text(self, text):
        print(text)
        self.main_window.statusbar.set_text(text)
    
    def on_files_selected(self, index):
        self.set_status_text("Viewing file: {}".format(self.filenames[index]))
    
    def selected_view(self):
        return self.views[self.tabwidget_images.current_index()]
    
    def on_fit(self):
        self.selected_view().fit_to_window()
    
    zoom_factor = 1.5
    
    def on_zoom_in(self):
        self.selected_view().zoom_by(self.zoom_factor)
    
    def on_zoom_out(self):
        self.selected_view().zoom_by(1.0 / self.zoom_factor)
    
    def on_rotate_left(self):
        self.selected_view().rotate_left()
    
    def on_rotate_right(self):
        self.selected_view().rotate_right()
    
    def on_about(self):
        self.about()
    
    def on_about_toolkit(self):
        self.about_toolkit()
    
    def on_exit(self):
        self.exit_app()


if __name__ == "__main__":
    if len(sys.argv) > 1:
        ImageViewer.filenames_to_open = sys.argv[1:]
    ImageViewer().run()



#!/usr/bin/env python3
"""
Module LISTVIEW -- UI Tree/List View Widgets
Sub-Package UI.BASE of Package PLIB3 -- Python UI Framework
Copyright (C) 2008-2022 by Peter A. Donis

Released under the GNU General Public License, Version 2
See the LICENSE and README files for more information
"""

import functools
import types

from plib.stdlib.coll import SortMixin

from plib.ui.defs import *
from plib.ui.base import helpers

from .app import PWidgetBase


class PListViewItemCols(helpers._PTreeHelperItemCols):
    """Exposes columns of a list view item as a sequence.
    """
    
    def _get_listviewitem(self):
        return self._helperitem
    
    listviewitem = property(_get_listviewitem)
    
    def _get_data(self, index):
        return self.listviewitem._get_col(index)
    
    def _set_data(self, index, value):
        self.listviewitem._set_col(index, value)


class PListViewHelperItem(helpers._PTreeHelperItem):
    # Common base class for items
    
    colsclass = PListViewItemCols
    
    def __init__(self, parent, index, data=None):
        self._parent = parent
        self._depth = 0
        self._listview = self._listview_from_parent(parent)
        helpers._PTreeHelperItem.__init__(self, parent, index, data)
    
    def _listview_from_parent(self, p):
        while not isinstance(p, PListViewMixin):
            p = p._parent
            self._depth += 1
        return p
    
    def _get_listview(self):
        return self._listview
    
    listview = property(_get_listview)
    
    def colcount(self):
        return self.listview.colcount()
    
    def _get_col(self, col):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def _set_col(self, col, value):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def indexes(self):
        parent = self._parent
        return parent.indexes() + (parent.index_of(self),)
    
    def current_item(self):
        item = self.listview.current_item()
        return item if item._value() in self else None
    
    def set_current_item(self, item):
        if item._value() in self:
            self.listview.set_current_item(item)


class ExpandMixin(object):
    # Mixin to handle common code for auto-expanding tree view
    
    def expand_all(self):
        for item in self.all_items():
            item.expand_all()


class PTreeViewItemBase(ExpandMixin, PListViewHelperItem):
    """Tree view item that looks like a sequence of 2-tuples.
    
    Each 2-tuple in the sequence is of the form
    (column-values, [list of child items]). The column-values
    is a list of strings; if the list view has only one column,
    the list has a single element.
    """
    
    def expand(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def expand_all(self):
        self.expand()
        ExpandMixin.expand_all(self)


class PListViewItemBase(PListViewHelperItem):
    """List view item that looks like a sequence of strings (columns)
    """
    
    def _value(self):
        cols, _ = PTreeViewItemBase._value(self)
        return cols


class PListViewLabelsBase(helpers._PHelperColLabels):
    
    def _get_listview(self):
        return self._helper
    
    listview = property(_get_listview)


class PListViewMixin(helpers._PTreeHelper):
    """Mixin class for PTreeViewBase and PListViewBase.
    
    Implements common behaviors.
    
    Includes convenience methods to handle common special cases.
    """
    
    signals = (
        SIGNAL_LISTVIEWSELECTED,
    )
    
    def __init__(self, parent, labels=None, data=None, header_font=None):
        # Some UI toolkits define a clear method but it
        # doesn't work reliably, so we insist on using
        # ours (note that we have to create the bound
        # method by hand)
        self.clear = types.MethodType(
            helpers._PTreeHelper.clear, self)
        self._parent = parent
        helpers._PTreeHelper.__init__(self, labels, data, header_font=header_font)
    
    def colcount(self):
        """Placeholder for derived classes to implement.
        """
        raise NotImplementedError
    
    def from_flat(self, items):
        """Canonicalize list of multi-column items with no children.
        """
        return ((item, []) for item in items)
    
    def from_list(self, items):
        """Canonicalize list of single-column items with no children.
        """
        return (([item], []) for item in items)
    
    def extend_flat(self, items):
        """Adds list of multi-column items to list.
        
        Assumes no children, but multiple columns.
        """
        self.extend(self.from_flat(items))
    
    def extend_list(self, items):
        """Adds list of strings to list.
        
        Assumes a single column list box with no children.
        """
        self.extend(self.from_list(items))


class PTreeViewBase(PWidgetBase, ExpandMixin, PListViewMixin):
    """Tree view that looks like a tree of list view items.
    """
    
    def __init__(self, manager, parent, labels=None, data=None, auto_expand=False,
                 font=None, header_font=None):
        
        PListViewMixin.__init__(self, parent, labels, data, header_font=header_font)
        PWidgetBase.__init__(self, manager, parent, font=font)
        if auto_expand:
            # Have to do this here because all children must already be added
            self.expand_all()
    
    def indexes(self):
        return ()
    
    def current_indexes(self):
        return self.current_item().indexes()
    
    def wrap_target(self, signal, target):
        if signal == SIGNAL_LISTVIEWSELECTED:
            @functools.wraps(target)
            def _wrapper(item):
                target(item.cols, item.children)
            return _wrapper
        return target


class PListViewBase(PWidgetBase, PListViewMixin):
    """List view that looks like a list of multi-column items.
    """
    
    def __init__(self, manager, parent, labels=None, data=None,
                 font=None, header_font=None):
        
        if data:
            data = self.from_flat(data)  # list views cannot have children for their items
        PListViewMixin.__init__(self, parent, labels, data, header_font=header_font)
        PWidgetBase.__init__(self, manager, parent, font=font)
    
    def wrap_target(self, signal, target):
        if signal == SIGNAL_LISTVIEWSELECTED:
            @functools.wraps(target)
            def _wrapper(item):
                target(item.cols)
            return _wrapper
        return target


class ListSortMixin(SortMixin):
    """Customized mixin class for sorted list view/list box.
    """
    
    # List view/list box keys are handled in a special way; we interpret
    # the key parameter as the function to apply to each column separately,
    # so we have to adjust it before using it
    
    def _init_key(self, key):
        # Construct key function that applies given key to each list column;
        # assume that our constructed function will be called on each item
        # in a sequence of (cols, children) tuples
        return (
            (lambda x: tuple(x[0])) if key is None else
            (lambda x: tuple(key(c) for c in x[0]))
        )


class ListItemSortMixin(ListSortMixin):
    """Customized mixin class for sorted list view/list box item.
    """
    
    def __init__(self, parent, index, data=None):
        # The data is a tuple (cols, children), we have to separate it
        # out so the inherited constructor initializes the cols but we
        # initialize the children
        if data is not None:
            cols, children = data
            data = (cols, [])
        else:
            children = None
        super(ListItemSortMixin, self).__init__(parent, index, data)
        self._init_seq(children, self.listview.key)


class PSortedListViewItemBase(ListItemSortMixin, PListViewItemBase):
    """Sorted list view item.
    """
    pass


class PSortedListViewBase(ListSortMixin, PListViewBase):
    """Customized mixin class for sorted list view.
    """
    
    def __init__(self, manager, parent, labels=None, data=None, key=None,
                 font=None, header_font=None):
        
        PListViewBase.__init__(self, manager, parent, labels=labels,
                               font=font, header_font=header_font)  # don't pass data here
        if data:
            data = self.from_flat(data)  # list views cannot have children for their items
        self._init_seq(data, key)  # sort items and add them here
